# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# flake8: noqa

from __future__ import annotations

import ctypes
from pathlib import Path
import platform
from typing import List, Tuple, Optional, Any, Union

from ansys.scadeone.core.common.exception import ScadeOneException
import ansys.scadeone.core.svc.simdata.core as core

###########################################################################
# wrap DLL functions

IS_ARCH_64 = ctypes.sizeof(ctypes.c_void_p) == 8

if IS_ARCH_64:
    root = Path(__file__).parents[3] / "libs"
    dll_name = (
        root / "win-x64" / "swan_sd.dll"
        if platform.system() == "Windows"
        else root / "linux-x64" / "libswan_sd.so"
    )
    dll_path = dll_name.absolute()
    SWAN_SD_CDLL = ctypes.CDLL(str(dll_path))
else:
    SWAN_SD_CDLL = None


def arch_64_error(*args):
    raise ScadeOneException("Simdata module is only compatible with 64 bit architecture")


def wrap_function(lib, func_name, argtypes, restype):
    """Wrapping ctypes functions"""

    if not IS_ARCH_64:
        return arch_64_error

    func = lib.__getattr__(func_name)
    func.restype = restype
    func.argtypes = argtypes
    return func


sdf_open_wrap = wrap_function(
    SWAN_SD_CDLL, "sdf_wopen", [ctypes.c_wchar_p, core.sdf_open_mode_t], core.sd_id_t
)
sdf_get_version_wrap = wrap_function(
    SWAN_SD_CDLL, "sdf_get_format_version", [core.sd_id_t], ctypes.c_char_p
)
sdf_dump_wrap = wrap_function(SWAN_SD_CDLL, "sdf_dump", [core.sd_id_t], core.sd_err_t)
sdf_close_wrap = wrap_function(SWAN_SD_CDLL, "sdf_close", [core.sd_id_t], core.sd_err_t)
sdt_get_n_all_types_wrap = wrap_function(
    SWAN_SD_CDLL, "sdt_get_n_all_types", [core.sd_id_t], core.sd_size_t
)
sdt_get_all_types_wrap = wrap_function(
    SWAN_SD_CDLL, "sdt_get_all_types", [core.sd_id_t, ctypes.POINTER(core.sd_id_t)], core.sd_err_t
)
sdt_get_class_wrap = wrap_function(SWAN_SD_CDLL, "sdt_get_class", [core.sd_id_t], core.sdt_class_t)
sdt_get_size_wrap = wrap_function(SWAN_SD_CDLL, "sdt_get_size", [core.sd_id_t], core.sd_size_t)
sdt_get_name_wrap = wrap_function(SWAN_SD_CDLL, "sdt_get_name", [core.sd_id_t], ctypes.c_char_p)
sdt_set_name_wrap = wrap_function(
    SWAN_SD_CDLL, "sdt_set_name", [core.sd_id_t, ctypes.c_char_p], core.sd_err_t
)

sdt_struct_create_wrap = wrap_function(
    SWAN_SD_CDLL, "sdt_struct_create", [core.sd_size_t], core.sd_id_t
)
sdt_struct_add_field_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdt_struct_add_field",
    [core.sd_id_t, ctypes.c_char_p, core.sd_size_t, core.sd_id_t],
    core.sd_err_t,
)
sdt_struct_get_n_fields_wrap = wrap_function(
    SWAN_SD_CDLL, "sdt_struct_get_n_fields", [core.sd_id_t], core.sd_size_t
)
sdt_struct_get_field_name_wrap = wrap_function(
    SWAN_SD_CDLL, "sdt_struct_get_field_name", [core.sd_id_t, core.sd_size_t], ctypes.c_char_p
)
sdt_struct_get_field_offset_wrap = wrap_function(
    SWAN_SD_CDLL, "sdt_struct_get_field_offset", [core.sd_id_t, core.sd_size_t], core.sd_size_t
)
sdt_struct_get_field_type_wrap = wrap_function(
    SWAN_SD_CDLL, "sdt_struct_get_field_type", [core.sd_id_t, core.sd_size_t], core.sd_id_t
)
sdt_array_create_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdt_array_create",
    [core.sd_id_t, core.sd_size_t, ctypes.POINTER(core.sd_size_t)],
    core.sd_id_t,
)
sdt_array_get_base_type_wrap = wrap_function(
    SWAN_SD_CDLL, "sdt_array_get_base_type", [core.sd_id_t], core.sd_id_t
)
sdt_array_get_n_dims_wrap = wrap_function(
    SWAN_SD_CDLL, "sdt_array_get_n_dims", [core.sd_id_t], core.sd_size_t
)
sdt_array_get_dims_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdt_array_get_dims",
    [core.sd_id_t, ctypes.POINTER(core.sd_size_t)],
    core.sd_err_t,
)
sdt_enum_create_wrap = wrap_function(SWAN_SD_CDLL, "sdt_enum_create", [core.sd_id_t], core.sd_id_t)
sdt_enum_add_value_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdt_enum_add_value",
    [core.sd_id_t, ctypes.c_char_p, ctypes.c_void_p],
    core.sd_err_t,
)
sdt_enum_get_base_type_wrap = wrap_function(
    SWAN_SD_CDLL, "sdt_enum_get_base_type", [core.sd_id_t], core.sd_id_t
)
sdt_enum_get_n_values_wrap = wrap_function(
    SWAN_SD_CDLL, "sdt_enum_get_n_values", [core.sd_id_t], core.sd_size_t
)
sdt_enum_get_value_name_wrap = wrap_function(
    SWAN_SD_CDLL, "sdt_enum_get_value_name", [core.sd_id_t, core.sd_size_t], ctypes.c_char_p
)
sdt_enum_get_value_value_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdt_enum_get_value_value",
    [core.sd_id_t, core.sd_size_t, ctypes.c_void_p],
    core.sd_err_t,
)
sdt_variant_create_wrap = wrap_function(
    SWAN_SD_CDLL, "sdt_variant_create", [core.sd_size_t], core.sd_id_t
)
sdt_variant_add_constructor_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdt_variant_add_constructor",
    [core.sd_id_t, ctypes.c_char_p, core.sd_size_t, core.sd_int32_t, core.sd_size_t, core.sd_id_t],
    core.sd_err_t,
)
sdt_variant_get_n_constructors_wrap = wrap_function(
    SWAN_SD_CDLL, "sdt_variant_get_n_constructors", [core.sd_id_t], core.sd_size_t
)
sdt_variant_get_constructor_name_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdt_variant_get_constructor_name",
    [core.sd_id_t, core.sd_size_t],
    ctypes.c_char_p,
)
sdt_variant_get_constructor_tag_offset_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdt_variant_get_constructor_tag_offset",
    [core.sd_id_t, core.sd_size_t],
    core.sd_size_t,
)
sdt_variant_get_constructor_tag_value_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdt_variant_get_constructor_tag_value",
    [core.sd_id_t, core.sd_size_t],
    core.sd_int32_t,
)
sdt_variant_get_constructor_value_offset_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdt_variant_get_constructor_value_offset",
    [core.sd_id_t, core.sd_size_t],
    core.sd_size_t,
)
sdt_variant_get_constructor_value_type_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdt_variant_get_constructor_value_type",
    [core.sd_id_t, core.sd_size_t],
    core.sd_id_t,
)

sdt_imported_create_wrap = wrap_function(
    SWAN_SD_CDLL, "sdt_imported_create", [core.sd_size_t], core.sd_id_t
)
sdt_vsize_imported_create_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdt_vsize_imported_create",
    [core.sd_size_t, core.sd_pfn_vsize_get_bytes_size_t, core.sd_pfn_vsize_to_bytes_t],
    core.sd_id_t,
)

sdt_imported_is_variable_size_wrap = wrap_function(
    SWAN_SD_CDLL, "sdt_imported_is_variable_size", [core.sd_id_t], core.sd_bool_t
)

sde_create_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sde_create",
    [core.sd_id_t, ctypes.c_char_p, core.sd_id_t, core.sde_kind_t],
    core.sd_id_t,
)

sde_find_wrap = wrap_function(
    SWAN_SD_CDLL, "sde_find", [core.sd_id_t, ctypes.c_char_p], core.sd_id_t
)
sde_find_part_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sde_find_part",
    [
        core.sd_id_t,
        ctypes.c_char_p,
        ctypes.POINTER(ctypes.c_char_p),
        ctypes.POINTER(core.sd_id_t),
        ctypes.POINTER(core.sd_size_t),
    ],
    core.sd_id_t,
)
sde_get_n_children_wrap = wrap_function(
    SWAN_SD_CDLL, "sde_get_n_children", [core.sd_id_t], core.sd_size_t
)
sde_get_child_wrap = wrap_function(
    SWAN_SD_CDLL, "sde_get_child", [core.sd_id_t, core.sd_size_t], core.sd_id_t
)
sde_get_children_wrap = wrap_function(
    SWAN_SD_CDLL, "sde_get_children", [core.sd_id_t, ctypes.POINTER(core.sd_id_t)], core.sd_err_t
)
sde_get_name_wrap = wrap_function(SWAN_SD_CDLL, "sde_get_name", [core.sd_id_t], ctypes.c_char_p)
sde_get_type_wrap = wrap_function(SWAN_SD_CDLL, "sde_get_type", [core.sd_id_t], core.sd_id_t)
sde_get_kind_wrap = wrap_function(SWAN_SD_CDLL, "sde_get_kind", [core.sd_id_t], core.sde_kind_t)
sde_set_name_wrap = wrap_function(
    SWAN_SD_CDLL, "sde_set_name", [core.sd_id_t, ctypes.c_char_p], core.sd_err_t
)
sde_set_type_wrap = wrap_function(
    SWAN_SD_CDLL, "sde_set_type", [core.sd_id_t, core.sd_id_t], core.sd_err_t
)
sde_set_kind_wrap = wrap_function(
    SWAN_SD_CDLL, "sde_set_kind", [core.sd_id_t, core.sde_kind_t], core.sd_err_t
)
sde_remove_wrap = wrap_function(SWAN_SD_CDLL, "sde_remove", [core.sd_id_t], core.sd_err_t)

sdd_append_raw_value_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_append_raw_value", [core.sd_id_t, ctypes.c_void_p], core.sd_err_t
)
sdd_append_value_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_append_value", [core.sd_id_t, core.sd_value_t], core.sd_err_t
)
sdd_append_sequence_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_append_sequence", [core.sd_id_t, core.sd_sequence_t], core.sd_err_t
)
sdd_value_iter_create_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_iter_create", [core.sd_id_t, core.sd_id_t], core.sd_value_iterator_t
)
sdd_value_iter_create_part_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdd_value_iter_create_part",
    [core.sd_id_t, ctypes.c_char_p, core.sd_id_t],
    core.sd_value_iterator_t,
)
sdd_value_iter_get_file_type_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_iter_get_file_type", [core.sd_value_iterator_t], core.sd_id_t
)
sdd_value_iter_get_mem_type_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_iter_get_mem_type", [core.sd_value_iterator_t], core.sd_id_t
)
sdd_value_iter_seek_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_iter_seek", [core.sd_value_iterator_t, core.sd_size_t], core.sd_err_t
)
sdd_value_iter_get_raw_value_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdd_value_iter_get_raw_value",
    [
        core.sd_value_iterator_t,
        ctypes.c_void_p,
        ctypes.POINTER(core.sd_bool_t),
        ctypes.POINTER(core.sd_bool_t),
    ],
    core.sd_err_t,
)
sdd_value_iter_get_raw_values_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdd_value_iter_get_raw_values",
    [core.sd_value_iterator_t, core.sd_size_t, ctypes.c_void_p, ctypes.POINTER(core.sd_bool_t)],
    core.sd_size_t,
)
sdd_value_iter_get_value_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_iter_get_value", [core.sd_value_iterator_t], core.sd_value_t
)
sdd_value_iter_set_raw_value_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdd_value_iter_set_raw_value",
    [core.sd_value_iterator_t, ctypes.c_void_p],
    core.sd_err_t,
)
sdd_value_iter_set_value_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdd_value_iter_set_value",
    [core.sd_value_iterator_t, core.sd_value_t],
    core.sd_err_t,
)
sdd_value_iter_clear_values_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_iter_clear_values", [core.sd_value_iterator_t], core.sd_err_t
)
sdd_value_iter_close_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_iter_close", [core.sd_value_iterator_t], None
)

# 'sequence iterator' object = an iterator on element sequences ('sequence' objects)
# TODO

sdd_value_create_none_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_create_none", [], core.sd_value_t
)
sdd_value_create_predef_char_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_create_predef_char8", [core.sd_uint8_t], core.sd_value_t
)
sdd_value_create_predef_bool_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_create_predef_bool", [core.sd_bool_t], core.sd_value_t
)
sdd_value_create_predef_int8_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_create_predef_int8", [core.sd_int8_t], core.sd_value_t
)
sdd_value_create_predef_int16_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_create_predef_int16", [core.sd_int16_t], core.sd_value_t
)
sdd_value_create_predef_int32_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_create_predef_int32", [core.sd_int32_t], core.sd_value_t
)
sdd_value_create_predef_int64_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_create_predef_int64", [core.sd_int64_t], core.sd_value_t
)
sdd_value_create_predef_uint8_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_create_predef_uint8", [core.sd_uint8_t], core.sd_value_t
)
sdd_value_create_predef_uint16_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_create_predef_uint16", [core.sd_uint16_t], core.sd_value_t
)
sdd_value_create_predef_uint32_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_create_predef_uint32", [core.sd_uint32_t], core.sd_value_t
)
sdd_value_create_predef_uint64_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_create_predef_uint64", [core.sd_uint64_t], core.sd_value_t
)
sdd_value_create_predef_float32_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_create_predef_float32", [core.sd_float32_t], core.sd_value_t
)
sdd_value_create_predef_float64_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_create_predef_float64", [core.sd_float64_t], core.sd_value_t
)
sdd_value_create_list_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdd_value_create_list",
    [core.sd_size_t, ctypes.POINTER(core.sd_value_t)],
    core.sd_value_t,
)
sdd_value_create_enum_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_create_enum", [ctypes.c_char_p], core.sd_value_t
)
sdd_value_create_variant_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_create_variant", [ctypes.c_char_p, core.sd_value_t], core.sd_value_t
)
sdd_value_create_imported_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_create_imported", [core.sd_size_t, ctypes.c_void_p], core.sd_value_t
)
sdd_value_close_wrap = wrap_function(SWAN_SD_CDLL, "sdd_value_close", [core.sd_value_t], None)
sdd_value_get_class_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_get_class", [core.sd_value_t], core.sdd_value_class_t
)
sdd_value_get_predef_type_id_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_get_predef_type_id", [core.sd_value_t], core.sd_id_t
)
sdd_value_get_predef_raw_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_get_predef_raw", [core.sd_value_t, ctypes.c_void_p], core.sd_err_t
)
sdd_value_get_predef_int8_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdd_value_get_predef_int8",
    [core.sd_value_t, ctypes.POINTER(core.sd_int8_t)],
    core.sd_err_t,
)
sdd_value_get_predef_int16_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdd_value_get_predef_int16",
    [core.sd_value_t, ctypes.POINTER(core.sd_int16_t)],
    core.sd_err_t,
)
sdd_value_get_predef_int32_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdd_value_get_predef_int32",
    [core.sd_value_t, ctypes.POINTER(core.sd_int32_t)],
    core.sd_err_t,
)
sdd_value_get_predef_int64_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdd_value_get_predef_int64",
    [core.sd_value_t, ctypes.POINTER(core.sd_int64_t)],
    core.sd_err_t,
)
sdd_value_get_predef_uint8_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdd_value_get_predef_uint8",
    [core.sd_value_t, ctypes.POINTER(core.sd_uint8_t)],
    core.sd_err_t,
)
sdd_value_get_predef_uint16_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdd_value_get_predef_uint16",
    [core.sd_value_t, ctypes.POINTER(core.sd_uint16_t)],
    core.sd_err_t,
)
sdd_value_get_predef_uint32_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdd_value_get_predef_uint32",
    [core.sd_value_t, ctypes.POINTER(core.sd_uint32_t)],
    core.sd_err_t,
)
sdd_value_get_predef_uint64_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdd_value_get_predef_uint64",
    [core.sd_value_t, ctypes.POINTER(core.sd_uint64_t)],
    core.sd_err_t,
)
sdd_value_get_predef_float32_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdd_value_get_predef_float32",
    [core.sd_value_t, ctypes.POINTER(core.sd_float32_t)],
    core.sd_err_t,
)
sdd_value_get_predef_float64_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdd_value_get_predef_float64",
    [core.sd_value_t, ctypes.POINTER(core.sd_float64_t)],
    core.sd_err_t,
)
sdd_value_get_enum_name_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_get_enum_name", [core.sd_value_t], ctypes.c_char_p
)
sdd_value_get_enum_value_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_get_enum_value", [core.sd_value_t], core.sd_value_t
)
sdd_value_get_list_n_values_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_get_list_n_values", [core.sd_value_t], core.sd_size_t
)
sdd_value_get_list_value_at_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_get_list_value_at", [core.sd_value_t, core.sd_size_t], core.sd_value_t
)
sdd_value_get_variant_name_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_get_variant_name", [core.sd_value_t], ctypes.c_char_p
)
sdd_value_get_variant_value_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_get_variant_value", [core.sd_value_t], core.sd_value_t
)
sdd_value_get_imported_size_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_get_imported_size", [core.sd_value_t], core.sd_size_t
)
sdd_value_get_imported_value_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_value_get_imported_value", [core.sd_value_t, ctypes.c_void_p], core.sd_err_t
)
sdd_sequence_create_nones_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_sequence_create_nones", [core.sd_size_t], core.sd_sequence_t
)
sdd_sequence_create_values_wrap = wrap_function(
    SWAN_SD_CDLL,
    "sdd_sequence_create_values",
    [core.sd_id_t, core.sd_size_t, ctypes.POINTER(core.sd_value_t), core.sd_size_t],
    core.sd_sequence_t,
)
sdd_sequence_close_wrap = wrap_function(
    SWAN_SD_CDLL, "sdd_sequence_close", [core.sd_sequence_t], None
)


###########################################################################
# file operations


def sdf_open(file_path: str, mode: core.FileOpenMode) -> int:
    b_file_path = ctypes.c_wchar_p(file_path)
    return sdf_open_wrap(b_file_path, mode)


def sdf_get_version(file_id: int) -> str:
    b_version = sdf_get_version_wrap(file_id)
    return b_version.decode("utf-8") if b_version is not None else ""


def sdf_dump(file_id: int) -> int:
    return sdf_dump_wrap(file_id)


def sdf_close(file_id: int) -> int:
    return sdf_close_wrap(file_id)


###########################################################################
# data types


# - common


def sdt_get_n_all_types(file_id: int) -> int:
    return sdt_get_n_all_types_wrap(file_id)


def sdt_get_all_types(file_id: int) -> List[int]:
    n_types = sdt_get_n_all_types(file_id)
    types_ids = (core.sd_id_t * n_types)()
    err = sdt_get_all_types_wrap(file_id, types_ids)
    return None if err != core.SD_ERR_NONE else list(types_ids)


def sdt_get_class(type_id: int) -> core.TypeClass:
    klass = sdt_get_class_wrap(type_id)
    return core.TypeClass(klass)


def sdt_get_size(type_id: int) -> int:
    return sdt_get_size_wrap(type_id)


def sdt_get_name(type_id: int) -> str:
    b_name = sdt_get_name_wrap(type_id)
    return b_name.decode("utf-8") if b_name is not None else ""


def sdt_set_name(type_id: int, name: str):
    b_name = name.encode("utf-8")
    return sdt_set_name_wrap(type_id, b_name)


# - struct


def sdt_struct_create(mem_size: int) -> int:
    return sdt_struct_create_wrap(mem_size)


def sdt_struct_add_field(struct_id: int, name: str, offset: int, type_id: int) -> int:
    b_name = name.encode("utf-8")
    return sdt_struct_add_field_wrap(struct_id, b_name, offset, type_id)


def sdt_struct_get_n_fields(struct_id: int) -> int:
    return sdt_struct_get_n_fields_wrap(struct_id)


def sdt_struct_get_field_name(struct_id: int, field_idx: int) -> str:
    b_name = sdt_struct_get_field_name_wrap(struct_id, field_idx)
    return b_name.decode("utf-8")


def sdt_struct_get_field_offset(struct_id: int, field_idx: int) -> int:
    return sdt_struct_get_field_offset_wrap(struct_id, field_idx)


def sdt_struct_get_field_type(struct_id: int, field_idx: int) -> int:
    return sdt_struct_get_field_type_wrap(struct_id, field_idx)


# - array


def sdt_array_create(base_type_id: int, n_dims: int, dims: List[int]):
    dims_array_type = core.sd_size_t * len(dims)
    dims_array = dims_array_type(*dims)
    return sdt_array_create_wrap(base_type_id, n_dims, dims_array)


def sdt_array_get_base_type(array_id: int) -> int:
    return sdt_array_get_base_type_wrap(array_id)


def sdt_array_get_n_dims(array_id: int) -> int:
    return sdt_array_get_n_dims_wrap(array_id)


def sdt_array_get_dims(array_id: int) -> List[int]:
    n_dims = sdt_array_get_n_dims(array_id)
    dims = (core.sd_size_t * n_dims)()
    err = sdt_array_get_dims_wrap(array_id, dims)
    return None if err != core.SD_ERR_NONE else list(dims)


# - enum


def sdt_enum_create(base_type_id: int) -> int:
    return sdt_enum_create_wrap(base_type_id)


def sdt_enum_add_value(enum_id: int, name: str, value: int) -> int:
    b_name = name.encode("utf-8")
    p_value = ctypes.addressof(ctypes.c_longlong(value))
    return sdt_enum_add_value_wrap(enum_id, b_name, p_value)


def sdt_enum_get_base_type(enum_id: int) -> int:
    return sdt_enum_get_base_type_wrap(enum_id)


def sdt_enum_get_n_values(enum_id: int) -> int:
    return sdt_enum_get_n_values_wrap(enum_id)


def sdt_enum_get_value_name(enum_id: int, value_idx: int) -> str:
    b_name = sdt_enum_get_value_name_wrap(enum_id, value_idx)
    return b_name.decode("utf-8")


def sdt_enum_get_value_value(enum_id: int, value_idx: int) -> int:
    v = ctypes.c_longlong(0)
    if sdt_enum_get_base_type(enum_id) == core.PredefinedType.UINT64.value:
        v = ctypes.c_ulonglong(0)
    err = sdt_enum_get_value_value_wrap(enum_id, value_idx, ctypes.byref(v))
    return None if err != core.SD_ERR_NONE else v.value


# - variant


def sdt_variant_create(mem_size: int) -> int:
    return sdt_variant_create_wrap(mem_size)


def sdt_variant_add_constructor(
    variant_id: int,
    name: str,
    tag_offset: int,
    tag_value: int,
    value_offset: int,
    value_type_id: int,
) -> int:
    b_name = name.encode("utf-8")
    return sdt_variant_add_constructor_wrap(
        variant_id, b_name, tag_offset, tag_value, value_offset, value_type_id
    )


def sdt_variant_get_n_constructors(variant_id: int) -> int:
    return sdt_variant_get_n_constructors_wrap(variant_id)


def sdt_variant_get_constructor_name(variant_id: int, constructor_idx: int) -> str:
    b_name = sdt_variant_get_constructor_name_wrap(variant_id, constructor_idx)
    return b_name.decode("utf-8")


def sdt_variant_get_constructor_tag_offset(variant_id: int, constructor_idx: int) -> int:
    return sdt_variant_get_constructor_tag_offset_wrap(variant_id, constructor_idx)


def sdt_variant_get_constructor_tag_value(variant_id: int, constructor_idx: int) -> int:
    return sdt_variant_get_constructor_tag_value_wrap(variant_id, constructor_idx)


def sdt_variant_get_constructor_value_offset(variant_id: int, constructor_idx: int) -> int:
    return sdt_variant_get_constructor_value_offset_wrap(variant_id, constructor_idx)


def sdt_variant_get_constructor_value_type(variant_id: int, constructor_idx: int) -> int:
    return sdt_variant_get_constructor_value_type_wrap(variant_id, constructor_idx)


# - imported


def sdt_imported_create(mem_size: int) -> int:
    return sdt_imported_create_wrap(mem_size)


to_avoid_gc = []


def sdt_vsize_imported_create(
    mem_size: int,
    pfn_get_bytes_size: core.sd_pfn_vsize_get_bytes_size_t,  # type: ignore
    pfn_vsize_to_bytes: core.sd_pfn_vsize_to_bytes_t,  # type: ignore
) -> int:
    f1 = core.sd_pfn_vsize_get_bytes_size_t(pfn_get_bytes_size)
    f2 = core.sd_pfn_vsize_to_bytes_t(pfn_vsize_to_bytes)
    to_avoid_gc.append([f1, f2])
    return sdt_vsize_imported_create_wrap(mem_size, f1, f2)


def sdt_imported_is_variable_size(imported_id: int) -> bool:
    return False if sdt_imported_is_variable_size_wrap(imported_id) == 0 else True


###########################################################################
# elements tree


# - create element


def sde_create(file_id: int, name: str, type_id: int, kind: core.SdeKind) -> int:
    b_name = name.encode("utf-8")
    return sde_create_wrap(file_id, b_name, type_id, kind.value)


# - find element in tree


def sde_find(file_or_elem_id: int, path: str) -> int:
    b_path = path.encode("utf-8")
    return sde_find_wrap(file_or_elem_id, b_path)


def sde_find_part(file_or_elem_id: int, path: str) -> Tuple[int, str, int, int]:
    b_path = path.encode("utf-8")
    part_path = ctypes.c_char_p()
    part_type_id = core.sd_id_t()
    part_offset = core.sd_size_t()
    elem_id = sde_find_part_wrap(
        file_or_elem_id,
        b_path,
        ctypes.byref(part_path),
        ctypes.byref(part_type_id),
        ctypes.byref(part_offset),
    )
    return (
        elem_id,
        None if part_path.value is None else part_path.value.decode("utf-8"),
        part_type_id.value,
        part_offset.value,
    )


# - access element properties / update / remove element


def sde_get_n_children(file_or_elem_id: int) -> int:
    return sde_get_n_children_wrap(file_or_elem_id)


def sde_get_child(file_or_elem_id: int, index: int) -> int:
    return sde_get_child_wrap(file_or_elem_id, index)


def sde_get_children(file_or_elem_id: int) -> List[int]:
    n_children = sde_get_n_children(file_or_elem_id)
    children_ids = (core.sd_id_t * n_children)()
    err = sde_get_children_wrap(file_or_elem_id, children_ids)
    return None if err != core.SD_ERR_NONE else list(children_ids)


def sde_get_name(elem_id: int) -> str:
    b_name = sde_get_name_wrap(elem_id)
    return b_name.decode("utf-8")


def sde_get_type(elem_id: int) -> int:
    return sde_get_type_wrap(elem_id)


def sde_get_kind(elem_id: int) -> core.SdeKind:
    kind = sde_get_kind_wrap(elem_id)
    return core.SdeKind(kind)


def sde_set_name(elem_id: int, name: str):
    b_name = name.encode("utf-8")
    return sde_set_name_wrap(elem_id, b_name)


def sde_set_type(elem_id: int, type_id: int) -> int:
    return sde_set_type_wrap(elem_id, type_id)


def sde_set_kind(elem_id: int, kind: core.SdeKind) -> int:
    return sde_set_kind_wrap(elem_id, kind.value)


def sde_remove(elem_id: int) -> int:
    return sde_remove_wrap(elem_id)


###########################################################################
# simulation data


# append value/sequence to element


def sdd_append_raw_value(elem_id: int, data: int) -> int:
    return sdd_append_raw_value_wrap(elem_id, data)


def sdd_append_value(elem_id: int, value: core.sd_value_t) -> int:
    return sdd_append_value_wrap(elem_id, value)


def sdd_append_sequence(elem_id: int, sequence: core.sd_sequence_t) -> int:
    return sdd_append_sequence_wrap(elem_id, sequence)


# 'value iterator' object = an iterator on element (or part of it) values ('value' objects / raw data)


def sdd_value_iter_create(
    elem_id: int, mem_type_id: int = core.SDT_NONE
) -> core.sd_value_iterator_t:
    return sdd_value_iter_create_wrap(elem_id, mem_type_id)


def sdd_value_iter_create_part(
    file_id: int, path: str, mem_type_id: int = core.SDT_NONE
) -> core.sd_value_iterator_t:
    b_path = path.encode("utf-8")
    return sdd_value_iter_create_part_wrap(file_id, b_path, mem_type_id)


def sdd_value_iter_get_file_type(iterator: core.sd_value_iterator_t) -> int:
    return sdd_value_iter_get_file_type_wrap(iterator)


def sdd_value_iter_get_mem_type(iterator: core.sd_value_iterator_t) -> int:
    return sdd_value_iter_get_mem_type_wrap(iterator)


def sdd_value_iter_seek(iterator: core.sd_value_iterator_t, n_values: int) -> int:
    return sdd_value_iter_seek_wrap(iterator, n_values)


def sdd_value_iter_get_raw_value(
    iterator: core.sd_value_iterator_t,
    data: ctypes.c_void_p,
    is_none: ctypes._POINTER[core.sd_bool_t],
    is_last: ctypes._POINTER[core.sd_bool_t],
) -> int:
    return sdd_value_iter_get_raw_value_wrap(iterator, data, is_none, is_last)


def sdd_value_iter_get_raw_values(
    iterator: core.sd_value_iterator_t,
    n_values: int,
    data: ctypes.c_void_p,
    is_none: ctypes._POINTER[core.sd_bool_t],
) -> int:
    return sdd_value_iter_get_raw_values_wrap(iterator, n_values, data, is_none)


def sdd_value_iter_get_value(iterator: core.sd_value_iterator_t) -> core.sd_value_t:
    return sdd_value_iter_get_value_wrap(iterator)


def sdd_value_iter_set_raw_value(iterator: core.sd_value_iterator_t, data: ctypes.c_void_p) -> int:
    return sdd_value_iter_set_raw_value_wrap(iterator, data)


def sdd_value_iter_set_value(iterator: core.sd_value_iterator_t, value: core.sd_value_t) -> int:
    return sdd_value_iter_set_value_wrap(iterator, value)


def sdd_value_iter_clear_values(iterator: core.sd_value_iterator_t) -> int:
    return sdd_value_iter_clear_values_wrap(iterator)


def sdd_value_iter_close(iterator: core.sd_value_iterator_t) -> None:
    sdd_value_iter_close_wrap(iterator)


# 'sequence iterator' object = an iterator on element sequences ('sequence' objects)
# TODO

# - 'value' constructors


def sdd_value_create_none() -> core.sd_value_t:
    return sdd_value_create_none_wrap()


def sdd_value_create_predef_char(value: str) -> Optional[core.sd_value_t]:
    if not isinstance(value, str) or len(value) != 1:
        return None
    return sdd_value_create_predef_char_wrap(core.sd_uint8_t(ord(value[0])))


def sdd_value_create_predef_bool(value: bool) -> Optional[core.sd_value_t]:
    if not isinstance(value, bool):
        return None
    return sdd_value_create_predef_bool_wrap(core.sd_bool_true if value else core.sd_bool_false)


def sdd_value_create_predef_int8(value: int) -> Optional[core.sd_value_t]:
    if not isinstance(value, int):
        return None
    return sdd_value_create_predef_int8_wrap(core.sd_int8_t(value))


def sdd_value_create_predef_int16(value: int) -> Optional[core.sd_value_t]:
    if not isinstance(value, int):
        return None
    return sdd_value_create_predef_int16_wrap(core.sd_int16_t(value))


def sdd_value_create_predef_int32(value: int) -> Optional[core.sd_value_t]:
    if not isinstance(value, int):
        return None
    return sdd_value_create_predef_int32_wrap(core.sd_int32_t(value))


def sdd_value_create_predef_int64(value: int) -> Optional[core.sd_value_t]:
    if not isinstance(value, int):
        return None
    return sdd_value_create_predef_int64_wrap(core.sd_int64_t(value))


def sdd_value_create_predef_uint8(value: int) -> Optional[core.sd_value_t]:
    if not isinstance(value, int):
        return None
    return sdd_value_create_predef_uint8_wrap(core.sd_uint8_t(value))


def sdd_value_create_predef_uint16(value: int) -> Optional[core.sd_value_t]:
    if not isinstance(value, int):
        return None
    return sdd_value_create_predef_uint16_wrap(core.sd_uint16_t(value))


def sdd_value_create_predef_uint32(value: int) -> Optional[core.sd_value_t]:
    if not isinstance(value, int):
        return None
    return sdd_value_create_predef_uint32_wrap(core.sd_uint32_t(value))


def sdd_value_create_predef_uint64(value: int) -> Optional[core.sd_value_t]:
    if not isinstance(value, int):
        return None
    return sdd_value_create_predef_uint64_wrap(core.sd_uint64_t(value))


def sdd_value_create_predef_float32(value: float) -> Optional[core.sd_value_t]:
    if not isinstance(value, float):
        return None
    return sdd_value_create_predef_float32_wrap(core.sd_float32_t(value))


def sdd_value_create_predef_float64(value: float) -> Optional[core.sd_value_t]:
    if not isinstance(value, float):
        return None
    return sdd_value_create_predef_float64_wrap(core.sd_float64_t(value))


def sdd_value_create_list(
    n_values: int, values: List[core.sd_value_t]
) -> Optional[core.sd_value_t]:
    if not isinstance(values, list):
        return None
    values_array_type = core.sd_value_t * len(values)
    values_array = values_array_type(*values)
    return sdd_value_create_list_wrap(n_values, values_array)


def sdd_value_create_enum(value: str) -> Optional[core.sd_value_t]:
    if not isinstance(value, str):
        return None
    return sdd_value_create_enum_wrap(value.encode("utf-8"))


def sdd_value_create_variant(
    name: str, value: Optional[core.sd_value_t]
) -> Optional[core.sd_value_t]:
    if not isinstance(name, str):
        return None
    return sdd_value_create_variant_wrap(
        name.encode("utf-8"), value if value is not None else core.sd_value_t(0)
    )


def sdd_value_create_imported(value: Any) -> Optional[core.sd_value_t]:
    c_t = value.__class__
    if not issubclass(c_t, (ctypes._SimpleCData, ctypes.Array, ctypes.Structure, ctypes.Union)):
        return None
    return sdd_value_create_imported_wrap(ctypes.sizeof(c_t), ctypes.addressof(value))


# - 'value' destructor


def sdd_value_close(value: core.sd_value_t) -> None:
    sdd_value_close_wrap(value)


# - 'value' accessors


def sdd_value_get_class(value: core.sd_value_t) -> core.DataClass:
    klass = sdd_value_get_class_wrap(value)
    return core.DataClass(klass)


def sdd_value_get_predef_type_id(value: core.sd_value_t) -> core.PredefinedType:
    kind = sdd_value_get_predef_type_id_wrap(value)
    return core.PredefinedType(kind)


def sdd_value_get_predef_raw(value: core.sd_value_t, data: ctypes.c_void_p) -> int:
    return sdd_value_get_predef_raw_wrap(value, data)


def sdd_value_get_predef_uint8(value: core.sd_value_t) -> Optional[core.sd_uint8_t]:
    data = core.sd_uint8_t(0)
    err = sdd_value_get_predef_uint8_wrap(value, ctypes.byref(data))
    return None if err != core.SD_ERR_NONE else data


def sdd_value_get_predef_uint16(value: core.sd_value_t) -> Optional[core.sd_uint16_t]:
    data = core.sd_uint16_t(0)
    err = sdd_value_get_predef_uint16_wrap(value, ctypes.byref(data))
    return None if err != core.SD_ERR_NONE else data


def sdd_value_get_predef_uint32(value: core.sd_value_t) -> Optional[core.sd_uint32_t]:
    data = core.sd_uint32_t(0)
    err = sdd_value_get_predef_uint32_wrap(value, ctypes.byref(data))
    return None if err != core.SD_ERR_NONE else data


def sdd_value_get_predef_uint64(value: core.sd_value_t) -> Optional[core.sd_uint64_t]:
    data = core.sd_uint64_t(0)
    err = sdd_value_get_predef_uint64_wrap(value, ctypes.byref(data))
    return None if err != core.SD_ERR_NONE else data


def sdd_value_get_predef_int8(value: core.sd_value_t) -> Optional[core.sd_int8_t]:
    data = core.sd_int8_t(0)
    err = sdd_value_get_predef_int8_wrap(value, ctypes.byref(data))
    return None if err != core.SD_ERR_NONE else data


def sdd_value_get_predef_int16(value: core.sd_value_t) -> Optional[core.sd_int16_t]:
    data = core.sd_int16_t(0)
    err = sdd_value_get_predef_int16_wrap(value, ctypes.byref(data))
    return None if err != core.SD_ERR_NONE else data


def sdd_value_get_predef_int32(value: core.sd_value_t) -> Optional[core.sd_int32_t]:
    data = core.sd_int32_t(0)
    err = sdd_value_get_predef_int32_wrap(value, ctypes.byref(data))
    return None if err != core.SD_ERR_NONE else data


def sdd_value_get_predef_int64(value: core.sd_value_t) -> Optional[core.sd_int64_t]:
    data = core.sd_int64_t(0)
    err = sdd_value_get_predef_int64_wrap(value, ctypes.byref(data))
    return None if err != core.SD_ERR_NONE else data


def sdd_value_get_predef_float32(value: core.sd_value_t) -> Optional[core.sd_float32_t]:
    data = core.sd_float32_t(0.0)
    err = sdd_value_get_predef_float32_wrap(value, ctypes.byref(data))
    return None if err != core.SD_ERR_NONE else data


def sdd_value_get_predef_float64(value: core.sd_value_t) -> Optional[core.sd_float64_t]:
    data = core.sd_float64_t(0.0)
    err = sdd_value_get_predef_float64_wrap(value, ctypes.byref(data))
    return None if err != core.SD_ERR_NONE else data


def sdd_value_get_enum_name(value: core.sd_value_t) -> str:
    b_name = sdd_value_get_enum_name_wrap(value)
    return b_name.decode("utf-8")


def sdd_value_get_enum_value(value: core.sd_value_t) -> core.sd_value_t:
    return sdd_value_get_enum_value_wrap(value)


def sdd_value_get_list_n_values(value: core.sd_value_t) -> int:
    return sdd_value_get_list_n_values_wrap(value)


def sdd_value_get_list_value_at(value: core.sd_value_t, index: int) -> core.sd_value_t:
    return sdd_value_get_list_value_at_wrap(value, index)


def sdd_value_get_variant_name(value: core.sd_value_t) -> str:
    b_name = sdd_value_get_variant_name_wrap(value)
    return b_name.decode("utf-8")


def sdd_value_get_variant_value(value: core.sd_value_t) -> core.sd_value_t:
    return sdd_value_get_variant_value_wrap(value)


def sdd_value_get_imported_size(value: core.sd_value_t) -> int:
    return sdd_value_get_imported_size_wrap(value)


def sdd_value_get_imported_value(value: core.sd_value_t):
    size = sdd_value_get_imported_size(value)
    bytes_data = (core.sd_byte_t * size)()
    err = sdd_value_get_imported_value_wrap(value, bytes_data)
    return None if err != core.SD_ERR_NONE else bytes_data


# 'sequence' object = (none | a list of values | a ramp [first..last]) * a repeat factor


# - 'sequence' constructors
def sdd_sequence_create_nones(count: int) -> Optional[core.sd_sequence_t]:
    return sdd_sequence_create_nones_wrap(count)


def sdd_sequence_create_values(
    type_id: int, count: int, values: List[core.sd_value_t], repeat_factor: int
) -> Optional[core.sd_sequence_t]:
    if not isinstance(values, list):
        return None
    values_array_type = core.sd_value_t * len(values)
    values_array = values_array_type(*values)
    return sdd_sequence_create_values_wrap(type_id, count, values_array, repeat_factor)


# - 'sequence' destructor


def sdd_sequence_close(sequence: core.sd_sequence_t) -> None:
    sdd_sequence_close_wrap(sequence)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# ===============================
# SSS Parser
# ===============================
# cSpell: ignore pyparsing, alphanums, printables, Hexa

from pathlib import Path
import re
from pyparsing import (
    DelimitedList,
    Forward,
    Group,
    Literal,
    Optional,
    ParseException,
    QuotedString,
    Regex,
    Suppress,
    Word,
    alphanums,
    nums,
    oneOf,
    printables,
    rest_of_line,
)
from .utils import ConverterLogger


class SSSParser:
    def __init__(self) -> None:
        self._ctors = None
        self._cmds = []
        self._defs = None

    def initialize(self, **sss_ctors) -> None:
        self._ctors = sss_ctors
        self._cmds = []
        self._bnf()

    def reset_cmd(self):
        self._cmds = []

    @staticmethod
    def merge_multiline(input_string: str) -> str:
        """Reunify multiline strings that are separated by a backslash"""
        buffer = []
        lines = input_string.splitlines()
        current_line = ""
        for line in lines:
            if line.endswith("\\"):
                current_line += line[:-1] + " "
            else:
                current_line += line
                buffer.append(current_line)
                current_line = ""
        lined_string = "\n".join(buffer)
        return lined_string

    def parse_file(self, file: Path):
        """Parse a SSS file and return the commands"""
        try:
            try:
                output_string = file.read_text(encoding="utf-8")
            except UnicodeDecodeError:
                output_string = file.read_text()
            lined_string = self.merge_multiline(output_string)
            self._defs.parse_string(lined_string, parse_all=True)
            cmds = self._cmds
            self._cmds = []
            return cmds

        except ParseException as e:
            ConverterLogger.exception(f"Error parsing file\n{file}\n{e.explain(depth=0)}")

    # Helper to parse an atom
    WS = {
        "\\n": "\n",
        "\\r": "\r",
        "\\t": "\t",
        "\\0": "\0",
    }

    @staticmethod
    def to_char(text: str):
        # Convert a string representing a char, potentially as '\\xhh", or ' \\n'
        # into the corresponding char
        # Text may be surrounded by single quotes or not
        if len(text) == 1:
            return text
        if text[0] == "'":
            text = text[1:-1]
        if text in SSSParser.WS:
            return SSSParser.WS[text]
        # handle '\'
        if text[0] != "\\":
            return text
        text = "0" + text[1:]
        char = "%c" % int(text, 16)
        return char

    HexaRe = re.compile(r"\\x[0-9a-fA-F]{2}")

    @staticmethod
    def string_to_chars(txt):
        # parse a string '"..."'. String must start with '"' and end with '"'
        def _parse_string(txt):
            if txt == '"':
                return []
            if txt[0:2] in ("\\n", "\\r", "\\t", "\\0"):
                char = txt[0:2]
                rest = txt[2:]
            elif SSSParser.HexaRe.match(txt[0:4]):
                char = txt[0:4]
                rest = txt[4:]
            elif txt[0] == "\\":
                char = txt[1]
                rest = txt[2:]
            else:
                char = txt[0]
                rest = txt[1:]
            return [SSSParser.to_char(char)] + _parse_string(rest)

        if txt[0] == '"':
            res = _parse_string(txt[1:])
            return res
        return []

    LCB, RCB = map(Suppress, "{}")

    @staticmethod
    def _bool_or_ident(txt: str):
        return (
            txt[0].lower() == "t"
            if txt in ("true", "false", "TRUE", "FALSE", "True", "False", "t", "f", "T", "F")
            else txt
        )

    @staticmethod
    def _parse_atom():
        # define the grammar
        LPAR, RPAR = map(Suppress, "()")
        int_value = Regex(r"[-+]?\d+").set_parse_action(lambda t: int(t[0]))
        hex_value = Regex(r"0x[0-9a-fA-F]+").set_parse_action(lambda t: int(t[0], 16))
        bin_value = Regex(r"0b[01]+").set_parse_action(lambda t: int(t[0], 2))
        oct_value = Regex(r"0o[0-7]+").set_parse_action(lambda t: int(t[0], 8))

        float_value = Regex(
            r"[-+]?\d*\.\d*(?:[eE][-+]?\d+)?|[-+]?\d+[eE][-+]?\d+"
        ).set_parse_action(lambda t: float(t[0]))
        float_special = oneOf("NaN qNaN sNaN +Inf -Inf").set_parse_action(lambda t: t[0])
        char_value = Literal("'''") ^ QuotedString("'", unquote_results=False)
        string_value = QuotedString('"', esc_char="\\", unquote_results=False)
        ident_or_kw = Regex(r"[\w+:]+")
        ident_or_kw.set_parse_action(lambda t: SSSParser._bool_or_ident(t[0]))

        value = Forward()
        atom = (  # order is important: hex/bin/oct before float before int
            hex_value
            | bin_value
            | oct_value
            | float_special
            | float_value
            | int_value
            | char_value
            | string_value
            | ident_or_kw
            | Literal("?")
            | Group(LPAR + DelimitedList(value) + RPAR)
        )
        value << Optional(atom).set_parse_action(lambda t: t if len(t) else ["<empty>"])
        return value ^ Group(SSSParser.LCB + value + SSSParser.RCB).set_parse_action(lambda t: t[0])

    def _bnf(self):
        # Build the parser

        variable = Regex(r"[\w\[\]\./:]+") ^ QuotedString(quote_char="{", end_quote_char="}")

        ident = Word(alphanums + "_")

        value = SSSParser._parse_atom()

        value.set_parse_action(lambda t: self._ctors["atom"](t.as_list()))

        set_def = (
            Literal("SSM::set")
            + variable.set_results_name("variable")
            + value.set_results_name("value")
        )

        cycle_def = Literal("SSM::cycle") + Optional(Word(nums).set_results_name("number"))

        check_image_def = Literal("SSM::check image") + rest_of_line

        complex_expr_def = (
            Word("{") + (Word("in") ^ Word("not") ^ (Word(alphanums) + Word("->"))) + rest_of_line
        ).set_results_name("check")

        check_complex_def = (
            Literal("SSM::check") + variable.set_results_name("variable") + complex_expr_def
        )

        check_arg = (
            Literal("sustain") + Literal("=") + Word(printables).set_results_name("sustain")
        ) ^ (Literal("real") + Literal("=") + Word(printables).set_results_name("real"))

        check_def = (
            Literal("SSM::check")
            + variable.set_results_name("variable")
            + value.set_results_name("value")
            + check_arg[...]
        )

        uncheck_def = Literal("SSM::uncheck") + variable

        alias_def = (
            Literal("SSM::alias")
            + Word(printables).set_results_name("alias")
            + variable.set_results_name("variable")
        )

        complex_alias_value_def = (
            Literal("SSM::alias_value") + ident.set_results_name("alias") + complex_expr_def
        )

        alias_value_def = (
            Literal("SSM::alias_value")
            + ident.set_results_name("alias")
            + value.set_results_name("value")
        )

        tolerance_def = Literal("SSM::set_tolerance") + rest_of_line.set_results_name("line")

        unsupported_def = Literal("SSM::") + rest_of_line

        comments_def = Literal("#") + rest_of_line

        def set_act(tokens) -> None:
            """Action for SSM::set command"""
            self._cmds.append(self._ctors["sss_set"](tokens.variable, tokens.value))

        def cycle_act(tokens) -> None:
            """Action for SSM::cycle command"""
            if tokens.number:
                number = tokens.number
            else:
                number = 1
            self._cmds.append(self._ctors["sss_cycle"](number))

        def check_act(tokens) -> None:
            """Action for SSM::check command"""
            if not tokens.sustain:
                sustain = 1
            elif str(tokens.sustain).lower() == "forever":
                sustain = 0
            else:
                sustain = int(tokens.sustain)
            self._cmds.append(
                self._ctors["sss_check"](tokens.variable, tokens.value, sustain, tokens.real)
            )

        def tolerance_act(tokens) -> None:
            """Action for SSM::tolerance command"""
            self._cmds.append(self._ctors["sss_tolerance"](tokens.line))

        def alias_act(tokens) -> None:
            """Action for SSM::alias command"""
            self._cmds.append(self._ctors["sss_alias"](tokens.alias, tokens.variable))

        def alias_value_act(tokens) -> None:
            """Action for SSM::alias_value command"""
            self._cmds.append(
                self._ctors["sss_alias_value"](tokens.alias, tokens.value, tokens.check)
            )

        def undefined_act(tokens) -> None:
            """Unsupported actions -> to be logged"""
            self._cmds.append(self._ctors["unsupported"](" ".join(tokens)))

        def no_act(tokens) -> None:
            """No action for comments"""
            pass

        comments_def.set_parse_action(no_act)

        set_def.set_parse_action(set_act)
        cycle_def.set_parse_action(cycle_act)
        check_def.set_parse_action(check_act)
        tolerance_def.set_parse_action(tolerance_act)
        alias_def.set_parse_action(alias_act)
        alias_value_def.set_parse_action(alias_value_act)
        complex_alias_value_def.set_parse_action(alias_value_act)

        check_image_def.set_parse_action(undefined_act)
        check_complex_def.set_parse_action(undefined_act)
        uncheck_def.set_parse_action(undefined_act)
        unsupported_def.set_parse_action(undefined_act)

        statement_def = (
            cycle_def
            | comments_def
            | check_image_def
            | check_complex_def
            | check_def
            | uncheck_def
            | tolerance_def
            | complex_alias_value_def
            | alias_value_def
            | alias_def
            | set_def
            | unsupported_def
        )

        self._defs = statement_def[...]



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Generator, Union, Optional, cast

from ..common.exception import ScadeOneException
from ..model import Model
from .common import Declaration, SwanItem
from .diagram import Diagram, SectionBlock
from .modules import (
    Module,
    GroupDeclarations,
    TypeDeclarations,
    ConstDeclarations,
    SensorDeclarations,
    ModuleInterface,
    ModuleBody,
)
from .operators import Operator, Signature
from .scopes import Scope, ScopeSection
from .scopesections import VarSection
from .variable import VarDecl


class ModuleNamespace:
    """Class to handle named objects defined in a module."""

    def __init__(self, module: Module, stop_search: bool = False) -> None:
        self._module = module
        # boolean to avoid infinite recursion between module interface and module body
        self._stop_search = stop_search

    def get_declaration(self, name: str) -> SwanItem:
        """Returns the declaration with the given name."""

        if not name:
            raise ScadeOneException("Name cannot be None or empty.")

        decl = ModuleNamespace._get_declaration(name, self._module)
        if decl is not None or self._stop_search:
            return decl
        model = cast(Model, self._module.model)
        if isinstance(self._module, ModuleBody):
            module_peer = model.get_module_interface(self._module.name.as_string)
        else:
            module_peer = model.get_module_body(self._module.name.as_string)
        if module_peer is not None:
            ns = ModuleNamespace(module_peer, stop_search=True)
            return ns.get_declaration(name)

    @staticmethod
    def _get_declaration(name: str, module: Module) -> SwanItem:
        # Returns declaration searching by name
        if name.find("::") == -1:
            for decl in module.declarations:
                found_decl = None
                if isinstance(decl, GroupDeclarations):
                    found_decl = ModuleNamespace._find_declaration(name, decl.groups)
                elif isinstance(decl, TypeDeclarations):
                    found_decl = ModuleNamespace._find_declaration(name, decl.types)
                elif isinstance(decl, ConstDeclarations):
                    found_decl = ModuleNamespace._find_declaration(name, decl.constants)
                elif isinstance(decl, SensorDeclarations):
                    found_decl = ModuleNamespace._find_declaration(name, decl.sensors)
                elif (
                    (isinstance(decl, Signature) or isinstance(decl, Operator)) and decl.id.value
                ) == name:
                    found_decl = decl
                if found_decl is not None:
                    return found_decl
            return None
        else:
            # look for path_id
            module_from_name = cast(Model, module.model).get_module_from_pathid(name, module)
            if module_from_name is None:
                raise ScadeOneException(f"Module not found: {name}.")
            module_ns = ModuleNamespace(module_from_name)
            return module_ns.get_declaration(name.split("::")[-1])

    @staticmethod
    def _find_declaration(
        name: str, declarations: Generator[Declaration, None, None]
    ) -> Optional[Declaration]:
        """Look for a declaration searching by name"""
        if name is None:
            raise ScadeOneException("Declaration name is None.")
        for decl in declarations:
            if decl.id is None:
                continue
            if decl.id.value == name:
                return decl
        return None


class ScopeNamespace:
    """Class to handle named objects defined in a scope.

    As scope can contain other scopes, each scope maintains a reference
    to its enclosing scope.
    """

    def __init__(self, scope: Union[Scope, ScopeSection]) -> None:
        self._scope = scope

    def get_declaration(self, name: str) -> SwanItem:
        """Returns the declaration with the given name."""
        from .scopes import Scope, ScopeSection

        if not name:
            raise ScadeOneException("Name cannot be None or empty.")
        if isinstance(self._scope, ScopeSection):
            return ScopeNamespace._get_section_declaration(name, self._scope)
        elif isinstance(self._scope, Scope):
            for section in self._scope.sections:
                return ScopeNamespace._get_section_declaration(name, section)

    @staticmethod
    def _get_section_declaration(name: str, section: ScopeSection) -> SwanItem:
        """Returns the section declaration with the given namespace."""
        if name.find("::") == -1:
            return ScopeNamespace._find_declaration(name, section)

        # look for path_id
        module_from_name = cast(Model, section.model).get_module_from_pathid(name, section.module)
        if module_from_name is None:
            raise ScadeOneException(f"Module not found: {name}.")
        module_ns = ModuleNamespace(module_from_name)
        return module_ns.get_declaration(name.split("::")[-1])

    @staticmethod
    def _find_declaration(name: str, item: SwanItem) -> SwanItem:
        # Look for declaration with the given name.

        if isinstance(item, ModuleBody):
            module_ns = ModuleNamespace(item)
            return module_ns.get_declaration(name)
        if isinstance(item, ModuleInterface):
            module_ns = ModuleNamespace(item)
            return module_ns.get_declaration(name)
        if isinstance(item, Operator):
            for input in item.inputs:
                if not isinstance(input, VarDecl):
                    raise ScadeOneException("Input is not a variable.")
                if input.id.value == name:
                    return input
            for output in item.outputs:
                if not isinstance(output, VarDecl):
                    raise ScadeOneException("Output is not a variable.")
                if output.id.value == name:
                    return output
        if isinstance(item, Scope):
            for section in item.sections:
                decl = ScopeNamespace._find_section_obj(name, section)
                if decl is not None:
                    return decl
        if isinstance(item, ScopeSection):
            decl = ScopeNamespace._find_section_obj(name, item)
            if decl is not None:
                return decl
        if item.owner is None:
            raise ScadeOneException(f"Item without owner: {type(item)}")
        return ScopeNamespace._find_declaration(name, item.owner)

    @staticmethod
    def _find_section_obj(name: str, section: ScopeSection) -> Declaration:
        """Returns the section object with the given name."""
        if isinstance(section, VarSection):
            for var_decl in section.var_decls:
                if var_decl.id.value == name:
                    return var_decl
        if isinstance(section, Diagram):
            for obj in filter(lambda sec_obj: isinstance(sec_obj, SectionBlock), section.objects):
                if not isinstance(obj.section, VarSection):
                    continue
                var_sec = ScopeNamespace._find_section_obj(name, obj.section)
                if var_sec is not None:
                    return var_sec
        return None



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from typing import Union
from ansys.scadeone.core.common.storage import JobStorage
from ansys.scadeone.core.interfaces import IProject
from ansys.scadeone.core.job import CodeGenerationJob, Job, JobType, SimulationJob, TestExecutionJob


class JobFactory:
    """Factory used to create the Job type corresponding to the given JobStorage"""

    def __new__(cls, *args, **kwargs) -> "JobFactory":
        if not cls._instance:
            cls._instance = super(JobFactory, cls).__new__(cls)
        return cls._instance

    @staticmethod
    def create_job(job_storage: JobStorage, sproj: IProject) -> Union[Job, None]:
        """Create a Job derivated object from a JobStorage and a project

        Parameters
        ----------
        job_storage : JobStorage
            Job storage
        sproj : Project
            Corresponding project

        Returns
        -------
        Job
            The new Job derivated object
        """
        json = job_storage.json
        name = json["Properties"]["Name"]
        job = None
        if json["Kind"] == str(JobType.CODEGEN):
            job = CodeGenerationJob(name, sproj, json, job_storage)
        elif json["Kind"] == str(JobType.SIMU):
            job = SimulationJob(name, sproj, json, job_storage)
        elif json["Kind"] == str(JobType.TESTEXEC):
            job = TestExecutionJob(name, sproj, json, job_storage)
        return job

    @staticmethod
    def new_job(kind: JobType, name: str, sproj: IProject) -> "Job":
        """Create a job from scratch

        Parameters
        ----------
        kind : JobType
            New job kind
        name : str
            New job name
        sproj : Project
            Corresponding Project. Necessary to determine sjob save location

        Returns
        -------
        Job
            Created job

        Raises
        ------
        ScadeOneException
            No valid job kind provided
        """
        if kind == JobType.CODEGEN:
            job = CodeGenerationJob(name, sproj)
        elif kind == JobType.SIMU:
            job = SimulationJob(name, sproj)
        elif kind == JobType.TESTEXEC:
            job = TestExecutionJob(name, sproj)
        else:
            from ansys.scadeone.core import ScadeOneException

            raise ScadeOneException("A valid job type must be provided at creation.")
        return job



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# for reference in Model
# IProject: app return type is 'scadeone.IScadeOne'
# see https://softwareengineering.stackexchange.com/questions/369146/how-to-avoid-bidirectional-class-and-module-dependencies  # noqa: E501
# The point is that ScadeOne and Project uses each other
# Alternative is to create an intermediate interfaces.py.
from pathlib import Path
from typing import Optional

from ansys.scadeone.core.common.storage import ProjectStorage

from typing import Union


class IProject:
    """Interface class"""

    @property
    def app(self) -> Optional["IScadeOne"]:
        return None

    @property
    def storage(self) -> Optional["ProjectStorage"]:
        return None

    @property
    def directory(self) -> Optional[Path]:
        return None

    def swan_sources(self, all: bool = False):
        return None


class IScadeOne:
    """Interface class"""

    @property
    def logger(self) -> None:
        return None

    @property
    def version(self) -> str:
        return ""

    @property
    def install_dir(self) -> Union[Path, None]:
        """Installation directory as given when creating the ScadeOne instance."""
        assert False

    def subst_in_path(self, _: str) -> str:
        assert False



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


# THIS MODULE CONTAINS THE CLASSES FOR TEST HARNESS AND TEST MODULE
# THIS IS NOT YET FULLY IMPLEMENTED

from typing import Union, List, Optional, Callable
from pathlib import Path

from ansys.scadeone.core.common.storage import SwanFile
from ansys.scadeone.core.swan.variable import VarDecl
from ansys.scadeone.core.swan.typedecl import Uint64Type, BoolType
from ansys.scadeone.core.swan.modules import Module, UseDirective, GlobalDeclaration
from ansys.scadeone.core.swan.scopes import Scope
import ansys.scadeone.core.swan.common as common


class Source(common.SwanItem):  # numpydoc ignore=PR01
    """Class representing a source in the harness."""

    def __init__(self, id: common.Identifier) -> None:
        super().__init__()
        self._id = id

    @property
    def id(self) -> common.Identifier:
        """Source identifier."""
        return self._id


class Oracle(common.SwanItem):  # numpydoc ignore=PR01
    """Class representing an oracle in the harness."""

    def __init__(self, id: common.Identifier) -> None:
        super().__init__()
        self._id = id

    @property
    def id(self) -> common.Identifier:
        """Oracle identifier."""
        return self._id


class SetSensorEquation(common.Equation):  # numpydoc ignore=PR01
    """Class representing a set sensor equation in the harness."""

    def __init__(self, sensor: common.PathIdentifier, value: common.Expression) -> None:
        super().__init__()
        self._sensor = sensor
        self._value = value

    @property
    def sensor(self) -> common.PathIdentifier:
        """Sensor identifier."""
        return self._sensor

    @property
    def value(self) -> common.Expression:
        """Value to set."""
        return self._value


class TestHarness(common.Declaration, common.PragmaBase, common.ModuleItem):  # numpydoc ignore=PR01
    """Test harness definition."""

    def __init__(
        self,
        id: common.Identifier,
        body: Union[Scope, common.Equation, None, Callable],
        pragmas: Optional[List[common.Pragma]] = None,
    ) -> None:
        common.Declaration.__init__(self, id)
        common.PragmaBase.__init__(self, pragmas)
        self._body = body
        self._inputs = [VarDecl(common.Identifier("_current_cycle"), type=Uint64Type())]
        self._outputs = [VarDecl(common.Identifier("_stop_condition"), type=BoolType())]

    @property
    def body(self) -> Union[Scope, common.Equation, None]:
        """Harness body: a scope, an equation, or None."""
        if isinstance(self._body, Callable):
            body = self._body(self)
            self._body = body
            self.set_owner(self, self._body)
        return self._body


class TestModule(Module):  # numpydoc ignore=PR01
    """Test module definition."""

    def __init__(
        self,
        name: common.PathIdentifier,
        use_directives: Optional[List[UseDirective]] = None,
        declarations: Optional[List[GlobalDeclaration]] = None,
    ) -> None:
        super().__init__(name, use_directives, declarations)

    @property
    def extension(self) -> str:
        """Returns module extension, with '.' included."""
        return ".swant"


def load_test_module(module: Path):
    """Load a test module from a file."""
    from ansys.scadeone.core.common.logger import LOGGER
    from ansys.scadeone.core.model.loader import SwanParser
    from ansys.scadeone.core.common.exception import ScadeOneException

    test = SwanFile(module)
    if not test.is_test:
        raise ScadeOneException(f"Model.load_source: unexpected file kind {test.path}.")
    ast = SwanParser(LOGGER).module_body(test)
    ast.source = str(test.path)
    return ast


if __name__ == "__main__":
    # Test the module
    test = load_test_module(
        Path(r"C:\Scade One\examples\QuadFlightControl\QuadFlightControl\assets\QuadTest.swant")
    )
    print(test)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
The class :py:class:`ScadeOneException` is used to raise an exception.
The message passed when raising the exception is passed to the
Scade One logger :py:class:`ansys.scadeone.core.common.logger.ScadeOneLogger`.
"""

from ansys.scadeone.core.common.logger import LOGGER


class ScadeOneException(Exception):
    """ScadeOne API Exception. When raising a :py:class:`ScadeOneException`
    it is automatically logged.
    See :py:class:`ansys.scadeone.core.common.logger.ScadeOneLogger` for more."""

    def __init__(self, message: str, exc_info: bool = False) -> None:
        super().__init__(message)
        LOGGER.exception(message, exc_info=exc_info)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


# cSpell: ignore levelname stpimporter

import argparse
import logging
from pathlib import Path

from ansys.scadeone.core.common.exception import LOGGER, ScadeOneException

from .stp_context import STPContext
from .swan_context import SwanContext
from .utils import ConverterLogger  # noqa: F401

logging.basicConfig(
    filename="stpimporter.log",
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    filemode="w",
)

LOGGER.logger = ConverterLogger.getLogger("stpimporter")


def cmd_parse():
    ConverterLogger.reset_counts()
    parser = argparse.ArgumentParser(
        prog="stpimporter", description="Convert test procedure to simulation data"
    )
    parser.add_argument("stp_path", type=str, help="STP file path")
    parser.add_argument("sproj_path", type=str, help="Associated Swan project path")
    parser.add_argument("-r", "--record_name", type=str, help="Specific record name")
    parser.add_argument("-o", "--output_directory", type=str, help="Use specific directory")
    parser.add_argument("-s", "--s_one_install", type=str, help="Scade One installation path")
    parser.add_argument("-v", "--verbose_sd", action="store_true", help="Print exported sd files")
    parser.add_argument(
        "-R", "--root", type=str, help="Specify root operator (override STP information)"
    )
    parser.add_argument(
        "-n", "--renamings", type=str, help="Renaming log file from Scade Suite importer"
    )
    parser.add_argument("--no_gc", action="store_true", help="Disable garbage collection")

    args = parser.parse_args()
    if not Path(args.stp_path).exists():
        parser.error(f"STP file does not exists: {args.stp_path}")
    if not Path(args.sproj_path).exists():
        parser.error(f"SPROJ file does not exists: {args.sproj_path}")
    if args.renamings is not None and not Path(args.renamings).exists():
        parser.error(f"Renaming file does not exists: {args.renamings}")

    swan_ctx = SwanContext(
        project_path=args.sproj_path, s_one_path=args.s_one_install, renamings=args.renamings
    )

    stp_ctx = STPContext(swan_ctx, args.root)

    stp_ctx.stp_path = args.stp_path
    stp_ctx.proj_path = args.sproj_path
    ConverterLogger.info(
        f"""Importing:
    STP: {args.stp_path}
    SPROJ: {args.sproj_path}
    Record: {args.record_name}
    Renamings: {args.renamings}
    Root: {args.root}
    S-One: {args.s_one_install}
"""
    )

    if args.verbose_sd:
        ConverterLogger.SD_VERBOSE = True

    if args.output_directory is not None:
        stp_ctx.output_dir = args.output_directory
    try:
        if args.record_name is not None:
            stp_ctx.load_record(args.record_name)
        else:
            stp_ctx.load_all_records()

        stp_ctx.start_all_converts(args.no_gc)
    except ScadeOneException as e:
        print(f"A problem has occurred during the conversion: {e}")
        return 1
    except Exception as e:
        LOGGER.exception("Failed with %s", e)
        print(f"Failed with {e}")
        return 1

    logging.info(
        "Summary: "
        f"""errors: {ConverterLogger.errors} """
        f"""warnings: {ConverterLogger.warnings}"""
    )
    if ConverterLogger.errors > 0:
        return 1
    if ConverterLogger.warnings > 0:
        return 2
    return 0



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
This module implements operator instances.
"""

from abc import ABC
from enum import Enum, auto
from typing import List, Optional, Union

import ansys.scadeone.core.swan.common as common
import ansys.scadeone.core.swan.scopes as scopes

from .expressions import ClockExpr, Group, GroupItem
from .variable import VarDecl


class OperatorBase(common.SwanItem, ABC):  # numpydoc ignore=PR01  # numpydoc ignore=PR01
    """Base class for: operator ::= prefix_op [[sizes]]."""

    def __init__(self, sizes: List[common.Expression]) -> None:
        common.SwanItem.__init__(self)
        self._sizes = sizes

    @property
    def sizes(self) -> List[common.Expression]:
        """Size parameters of call."""
        return self._sizes


class PathIdOpCall(OperatorBase, common.PragmaBase):  # numpydoc ignore=PR01
    """Call to user-defined operator: operator ::= path_id [[sizes]]."""

    def __init__(
        self,
        path_id: common.PathIdentifier,
        sizes: List[common.Expression],
        pragmas: List[common.Pragma],
    ) -> None:
        OperatorBase.__init__(self, sizes)
        common.PragmaBase.__init__(self, pragmas)
        self._path_id = path_id

    @property
    def path_id(self) -> common.PathIdentifier:
        """Operator path."""
        return self._path_id


class PrefixPrimitiveKind(Enum):
    """Prefix primitive kind: reverse, transpose, pack, and flatten."""

    # pylint disable=invalid-name

    #: **flatten** array operation.
    Flatten = auto()
    #: **pack** array operation.
    Pack = auto()
    #: **reverse** array operation.
    Reverse = auto()
    #: **transpose** array operation.
    Transpose = auto()

    @staticmethod
    def to_str(value: "PrefixPrimitiveKind") -> str:
        return value.name.lower()


class PrefixPrimitive(OperatorBase):  # numpydoc ignore=PR01
    """
    Call to primitive operator: operator ::= *prefix_primitive* [[sizes]]
    with *prefix_primitive*:
    **flatten**,
    **pack**,
    **reverse**,
    operators."""

    def __init__(self, kind: PrefixPrimitiveKind, sizes: List[common.Expression]) -> None:
        super().__init__(sizes)
        self._kind = kind

    @property
    def kind(self) -> PrefixPrimitiveKind:
        """Primitive kind."""
        return self._kind


class Transpose(PrefixPrimitive):  # numpydoc ignore=PR01
    """Transpose operator.

    Parameters are a list of integer, but could be a
    single string if the indices are syntactically incorrect."""

    def __init__(self, params: Union[List[str], str], sizes: List[common.Expression]) -> None:
        super().__init__(PrefixPrimitiveKind.Transpose, sizes)
        self._params = params
        self._is_valid = isinstance(params, list)

    @property
    def params(self) -> Union[List[str], str]:
        """Transpose indices as list of str is transpose indices
        are valid, else the invalid string of indices."""
        return self._params

    @property
    def is_valid(self) -> bool:
        """True when indices are a valid list, else one has a string."""
        return self._is_valid


class OperatorExpression(common.SwanItem, ABC):  # numpydoc ignore=PR01
    """Base class for *op_expr*."""

    def __init__(self) -> None:
        common.SwanItem.__init__(self)


class PrefixOperatorExpression(OperatorBase):  # numpydoc ignore=PR01
    """Call to *op_expr*: operator ::= (*op_expr*) [[sizes]]."""

    def __init__(self, op_expr: OperatorExpression, sizes: List[common.Expression]) -> None:
        super().__init__(sizes)
        self._op_expr = op_expr
        self._is_text = False
        self._is_op_expr = False

    @property
    def op_expr(self) -> OperatorExpression:
        """Operator expression."""
        return self._op_expr

    @property
    def is_text(self) -> bool:
        """Return True if *op_expr* is read from {text% ...%text}."""
        return self._is_text

    @is_text.setter
    def is_text(self, is_text: bool):
        self._is_text = is_text

    @property
    def is_op_expr(self) -> bool:
        """Return True if *op_expr* is read from {op_expr% ...%op_expr}."""
        return self._is_op_expr

    @is_op_expr.setter
    def is_op_expr(self, is_op_expr: bool):
        self._is_op_expr = is_op_expr


class IteratorKind(Enum):
    """Iterators kind: map, fold, mapfold, mapi, foldi, mapfoldi."""

    # pylint disable=invalid-name,no-else-return

    #: **map** iterator.
    Map = auto()
    #: **fold** iterator.
    Fold = auto()
    #: **mapfold** iterator.
    Mapfold = auto()
    #: **mapi** iterator.
    Mapi = auto()
    #: **foldi** iterator.
    Foldi = auto()
    #: **mapfoldi** iterator.
    Mapfoldi = auto()

    @staticmethod
    def to_str(value: "IteratorKind") -> str:
        if value == IteratorKind.Map:
            return "map"
        elif value == IteratorKind.Fold:
            return "fold"
        elif value == IteratorKind.Mapfold:
            return "mapfold"
        elif value == IteratorKind.Mapi:
            return "mapi"
        elif value == IteratorKind.Foldi:
            return "foldi"
        elif value == IteratorKind.Mapfoldi:
            return "mapfoldi"


class Iterator(OperatorExpression):  # numpydoc ignore=PR01
    """Iterators: map, fold, mapfold, mapi, foldi, mapfoldi."""

    def __init__(self, kind: IteratorKind, operator: OperatorBase) -> None:
        super().__init__()
        self._kind = kind
        self._operator = operator

    @property
    def kind(self) -> IteratorKind:
        """Iterator kind."""
        return self._kind

    @property
    def operator(self) -> OperatorBase:
        """Iterated operator."""
        return self._operator


class ActivateClock(OperatorExpression):  # numpydoc ignore=PR01
    """**activate** *operator* **every** *clock_expr*"""

    def __init__(self, operator: OperatorBase, clock: ClockExpr) -> None:
        super().__init__()
        self._operator = operator
        self._clock = clock

    @property
    def operator(self) -> OperatorBase:
        """Operator under activation"""
        return self._operator

    @property
    def clock(self) -> ClockExpr:
        """Activation clock expression"""
        return self._clock


class ActivateEvery(OperatorExpression):  # numpydoc ignore=PR01
    """Higher-order activate expression: **activate** *operator*
    **every** *expr* (( **last**| **default** )) *expr*."""

    def __init__(
        self,
        operator: OperatorBase,
        condition: common.Expression,
        is_last: bool,
        expr: common.Expression,
    ) -> None:
        super().__init__()
        self._operator = operator
        self._condition = condition
        self._is_last = is_last
        self._expr = expr

    @property
    def operator(self) -> OperatorBase:
        """Operator under activation."""
        return self._operator

    @property
    def condition(self) -> common.Expression:
        """Activation condition."""
        return self._condition

    @property
    def is_last(self) -> bool:
        """Returns true when **last** is set, false when **default** is set."""
        return self._is_last

    @property
    def expr(self) -> common.Expression:
        """Activation default/last expression."""
        return self._expr


class Restart(OperatorExpression):  # numpydoc ignore=PR01
    """Higher-order restart expression: **restart** *operator* **every** *expr*."""

    def __init__(self, operator: OperatorBase, condition: common.Expression) -> None:
        super().__init__()
        self._operator = operator
        self._condition = condition

    @property
    def operator(self) -> OperatorBase:
        """Operator under activation."""
        return self._operator

    @property
    def condition(self) -> common.Expression:
        """Activation condition."""
        return self._condition


class OptGroupItem(common.SwanItem):  # numpydoc ignore=PR01
    """Optional group item: *opt_group_item* ::= _ | *group_item*."""

    def __init__(self, item: Optional[GroupItem] = None) -> None:
        super().__init__()
        self._item = item

    @property
    def is_underscore(self) -> bool:
        """True when group item is '_'."""
        return self._item is None

    @property
    def item(self) -> Union[GroupItem, None]:
        """Returns the group item, either a GroupItem or None."""
        return self._item


class Partial(OperatorExpression):  # numpydoc ignore=PR01
    r"Partial operator expression: *operator* \ *partial_group*."

    def __init__(self, operator: OperatorBase, partial_group: List[OptGroupItem]) -> None:
        super().__init__()
        self._operator = operator
        self._partial_group = partial_group

    @property
    def operator(self) -> OperatorBase:
        """Called operator."""
        return self._operator

    @property
    def partial_group(self) -> List[OptGroupItem]:
        """Returns the partial group items."""
        return self._partial_group


class NaryOp(Enum):
    """N-ary operators."""

    # pylint disable=invalid-name,no-else-return
    #: (+) N-ary addition.
    Plus = auto()
    #: (*) N-ary multiplication.
    Mult = auto()
    #: (**land**) N-ary bitwise and.
    Land = auto()
    #: (**lor**) N-ary bitwise or.
    Lor = auto()
    #: (**and**) N-ary logical and.
    And = auto()
    #: (**or**) N-ary logical or.
    Or = auto()
    #: (**xor**) N-ary exclusive or.
    Xor = auto()
    #: (**@**) N-ary array concat.
    Concat = auto()

    @staticmethod
    def to_str(value: "NaryOp"):
        if value == NaryOp.Plus:
            return "+"
        elif value == NaryOp.Mult:
            return "*"
        elif value == NaryOp.Land:
            return "land"
        elif value == NaryOp.Lor:
            return "lor"
        elif value == NaryOp.And:
            return "and"
        elif value == NaryOp.Or:
            return "or"
        elif value == NaryOp.Xor:
            return "xor"
        elif value == NaryOp.Concat:
            return "@"


class NAryOperator(OperatorExpression):  # numpydoc ignore=PR01
    """N-ary operators: '+' | '*' | '@' | **and** | **or** | **xor** | **land** | **lor**."""

    def __init__(self, operator: NaryOp) -> None:
        super().__init__()
        self._operator = operator

    @property
    def operator(self) -> NaryOp:
        """N-ary operator."""
        return self._operator


class AnonymousOperatorWithExpression(OperatorExpression):  # numpydoc ignore=PR01
    """Anonymous operator expression:
    ((**node|function**)) id {{ , id }} *scope_sections* => *expr*."""

    def __init__(
        self,
        is_node: bool,
        params: List[common.Identifier],
        sections: List[scopes.ScopeSection],
        expr: common.Expression,
    ) -> None:
        super().__init__()
        self._is_node = is_node
        self._params = params
        self._sections = sections
        self._expr = expr

    @property
    def is_node(self) -> bool:
        """True when anonymous operator is a node, else a function."""
        return self._is_node

    @property
    def params(self) -> List[common.Identifier]:
        """Anonymous operator parameters list."""
        return self._params

    @property
    def sections(self) -> List[scopes.ScopeSection]:
        """Scope sections list."""
        return self._sections

    @property
    def expr(self) -> common.Expression:
        """Anonymous operator body."""
        return self._expr


class AnonymousOperatorWithDataDefinition(OperatorExpression):  # numpydoc ignore=PR01
    """Anonymous operator expression:
    ((**node|function**)) *params* **returns** *params* *data_def*."""

    def __init__(
        self,
        is_node: bool,
        inputs: List[VarDecl],
        outputs: List[VarDecl],
        data_def: Union[common.Equation, scopes.Scope],
    ) -> None:
        super().__init__()
        self._is_node = is_node
        self._inputs = inputs
        self._outputs = outputs
        self._data_def = data_def

    @property
    def is_node(self) -> bool:
        """True when anonymous operator is a node, else a function."""
        return self._is_node

    @property
    def inputs(self) -> List[VarDecl]:
        """Anonymous operator input list."""
        return self._inputs

    @property
    def outputs(self) -> List[VarDecl]:
        """Anonymous operator output list."""
        return self._outputs

    @property
    def data_def(self) -> Union[common.Equation, scopes.Scope]:
        """Anonymous operator data definition, either an equation or a scope."""
        return self._data_def


class OperatorInstance(common.Expression):  # numpydoc ignore=PR01
    """Operator instance call:

    *expr* := *operator_instance* ( *group* )

    *operator_instance* ::= *operator* [[ luid ]]"""

    def __init__(
        self, operator: OperatorBase, params: Group, luid: Optional[common.Luid] = None
    ) -> None:
        super().__init__()
        self._operator = operator
        self._params = params
        self._luid = luid

    @property
    def operator(self) -> OperatorBase:
        """Called operator."""
        return self._operator

    @property
    def params(self) -> Group:
        """Call parameters."""
        return self._params

    @property
    def luid(self) -> Union[common.Luid, None]:
        """Optional luid."""
        return self._luid


# =============================================
# Protected Items
# =============================================


class ProtectedOpExpr(OperatorExpression, common.ProtectedItem):  # numpydoc ignore=PR01
    """Protected operator expression,
    i.e., saved as string if syntactically incorrect."""

    def __init__(self, data: str, markup: str) -> None:
        common.ProtectedItem.__init__(self, data, markup)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from pathlib import Path
from collections import namedtuple
from platformdirs import PlatformDirs
import re

# Version must be directly defined for flit. No computation, else flit will fails
__version__ = "0.7.1"

m = re.match(
    r"""(?P<M>\d+)\.(?P<m>\d+)\.(?P<p>\d+)   # Major, minor, patch
        (?:\+(?P<b>\d+))?                    # Build (optional)
        (?:(?P<pr>\D.*))?                    # Pre-release (optional)
    """,
    __version__,
    re.X,
)

# version as a named tuple
Version = namedtuple("Version", ["major", "minor", "patch", "build", "pre_release"])

version_info = (
    Version(
        m.group("M"),
        m.group("m"),
        m.group("p"),
        m.group("b") if m.group("b") else "",
        m.group("pr") is not None,
    )
    if m
    else Version(0, 0, 0, 0, False)
)

# full version as a string
full_version = ".".join([version_info.major, version_info.minor, version_info.patch])
if version_info.build:
    full_version += f"+{version_info.build}"
if version_info.pre_release:
    full_version += " - Prerelease"

PYSCADEONE_DIR = Path(__file__).parent
PLATFORM_DIRS = PlatformDirs("PyScadeOne", "Ansys")

# pylint: disable=wrong-import-position
from .scadeone import ScadeOne  # noqa as we export name
from .common.exception import ScadeOneException  # noqa as we export name
from .common.storage import ProjectFile, SwanFile  # noqa as we export name



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# This module is used to create an FMI 2.0 co-simulation or model-exchange FMU
# from a Scade One model.
# The discrete operator is executed periodically.
#
# It generates C file needed by the runtime
# and the XML description of the FMU (modelDescription.xml).
# The FMU can then be created by compiling all source files with the corresponding runtime
# into a DLL and then packing it with the generated modelDescription.xml file.

# cSpell: ignore elems oper codegen mvars outdir newl addindent lshlwapi

from abc import ABC, abstractmethod
import os
from pathlib import Path
import re
import shutil
import subprocess
from typing import List, Optional, Tuple, Union
import uuid
import xml.dom.minidom as D
import zipfile

import jinja2

from ansys.scadeone.core.common.exception import ScadeOneException
from ansys.scadeone.core.common.logger import LOGGER

from ansys.scadeone.core.project import Project
import ansys.scadeone.core.svc.generated_code as GC

script_dir = Path(__file__).parent

CYCLE_FUNCTION_RETURN = "cycle_function_return"

# Model variables


class ModelVar:
    """
    Class representing a scalar component of a variable appearing
    in the interface of the FMU.
    """

    def __init__(
        self, fmu: "FMU_Export", sc_path: str, c_path: str, type_elem: dict, var_kind: str
    ) -> None:
        self._fmu = fmu
        self._c_path = c_path
        self._sc_path = sc_path
        self._type_elem = type_elem
        self._var_kind = var_kind
        self._type_kind = None
        self._xml_description = None
        self._xml_name = None
        self._direction = None

    @property
    def direction(self) -> str:
        """Return the direction ('input' or 'output') of the variable"""
        if self._direction is None:
            self._direction = "output" if self._var_kind == "output" else "input"
        return self._direction

    @property
    def type_name(self) -> str:
        """Return the generated type name of the variable"""
        return self._type_elem["name"]

    @property
    def type_kind(self) -> str:
        """Return the name of the FMU type corresponding to generated type of a given variable."""

        def fmu_ty_of_scade_ty(ty: dict) -> str:
            type_name = ty["name"]
            category = ty["category"]

            if category == "predefined_type":
                if type_name in ("swan_float32", "swan_float64"):
                    fmu_ty = "Real"
                elif type_name in (
                    "swan_int8",
                    "swan_int16",
                    "swan_int32",
                    "swan_int64",
                    "swan_uint8",
                    "swan_uint16",
                    "swan_uint32",
                    "swan_uint64",
                    "swan_char",
                    "swan_size",
                ):
                    fmu_ty = "Integer"
                elif type_name == "swan_bool":
                    fmu_ty = "Boolean"
                else:
                    raise ScadeOneException(
                        f"FMU_Export: Variable {self._sc_path} of type {type_name}: "
                        f"type not supported"
                    )
            elif category == "enum":
                fmu_ty = "Integer"
            elif category == "typedef":
                raise ScadeOneException(
                    f"FMU_Export: Variable {self._sc_path} of type {type_name}: "
                    f"imported types are not supported"
                )
            else:
                raise ScadeOneException(
                    f"FMU_Export: Variable {self._sc_path} of type {type_name}: "
                    f"category {category} is not supported"  # noqa
                )
            return fmu_ty

        if self._type_kind is None:
            self._type_kind = fmu_ty_of_scade_ty(self._type_elem)
        return self._type_kind

    @property
    def oper_path(self) -> str:
        """Return the Scade One path for the exported operator owning the variable."""
        return self._fmu.oper_path + "/"

    @property
    def xml_description(self) -> str:
        """Return the description of the variable used by the FMI model description"""
        if self._xml_description is None:
            if self._var_kind == "sensor":
                self._xml_description = self._sc_path
            else:
                self._xml_description = self.oper_path + self._sc_path
        return self._xml_description

    @property
    def xml_name(self) -> str:
        """Return the name of the variable used by the FMI model description"""

        def _replace_brackets(match):
            # Find all digits in the match and join them with commas
            numbers = re.findall(r"\d+", match.group(0))
            return f"[{','.join(numbers)}]"

        if self._xml_name is None:
            if self._var_kind == "sensor":
                self._xml_name = self._c_path
            else:
                self._xml_name = self._sc_path
            # update multidimensional arrays representation: [i][j] => [i,j]
            self._xml_name = re.sub(r"(\[\d+]){2,}", _replace_brackets, self._xml_name)
        return self._xml_name

    def get_default_value(self, xml=True) -> str:
        """
        Return the default value corresponding to FMU type of a given variable.
        Expected types are 'Real', 'Integer' and 'Boolean'.
        """
        if self.type_kind == "Real":
            return "0.0"
        elif self.type_kind == "Integer":
            return "0"
        elif xml:
            return "false"
        else:
            return "fmi2False"

    @staticmethod
    def paths_of_param(sc_path: str, c_path: str, code_type: dict) -> List[Tuple[str, str, dict]]:
        """
        Return the list of paths of scalar variables corresponding
        to the variable named `name` of type `ty` (of type `mapping.C.Type`).
        """
        var_list = []

        if code_type["category"] == "array":
            base_type = code_type["elements"]["base_type"]
            for i in range(0, code_type["elements"]["size"]):
                var_list.extend(
                    ModelVar.paths_of_param(f"{sc_path}[{i}]", f"{c_path}[{i}]", base_type)
                )
        elif code_type["category"] == "struct":
            for f in code_type["elements"]:
                var_list.extend(
                    ModelVar.paths_of_param(
                        f"{sc_path}.{f['name']}", f"{c_path}.{f['name']}", f["type"]
                    )
                )
        elif code_type["category"] == "union":
            for f in code_type["elements"]:
                var_list.extend(
                    ModelVar.paths_of_param(
                        f"{sc_path}.{f['name']}", f"{c_path}.{f['name']}", f["type"]
                    )
                )
        else:
            var_list.append((sc_path, c_path, code_type))
        return var_list

    @staticmethod
    def model_vars_of_param(
        fmu: "FMU_Export", v: Union[GC.ModelVariable, GC.ModelSensor], var_kind: str
    ) -> List["ModelVar"]:
        """
        Return the list of variables corresponding
        to the given model variable (input or output) or sensor.
        """
        if isinstance(v, GC.ModelVariable):
            sc_path = v.full_name(".")
        else:
            sc_path = v.path
        c_path = v.code_name
        if c_path == "__no_name__":
            # output of cycle function
            c_path = CYCLE_FUNCTION_RETURN
        code_type = v.code_type
        paths = ModelVar.paths_of_param(sc_path, c_path, code_type)
        return [ModelVar(fmu, sc_path, c_path, ty, var_kind) for sc_path, c_path, ty in paths]

    def append_xml(self, parent: D.Element) -> None:
        """
        Adds the XML element describing this model variable in FMI 2.0
        as a child of `parent`.
        """
        d = self._fmu.create_xml_child("ScalarVariable", parent)
        d.setAttribute("causality", self.direction)
        d.setAttribute("description", self.xml_description)
        d.setAttribute("name", self.xml_name)
        fmu_ty = self.type_kind
        d.setAttribute("valueReference", self._fmu.get_next_value_reference(fmu_ty))
        if fmu_ty == "Real":
            d.setAttribute("variability", "continuous")
        else:
            d.setAttribute("variability", "discrete")
        if self.direction == "output":
            d.setAttribute("initial", "calculated")
        ty = self._fmu.create_xml_child(fmu_ty, d)
        if self.direction == "input":
            ty.setAttribute("start", self.get_default_value())

    def get_context_path(self, fmu_type: str = "") -> str:
        # Returns the C expression computing the offset of this model variable
        # in the runtime state.
        cast = "" if fmu_type == "" else f"(fmi2{fmu_type})"
        path = "" if self._var_kind == "sensor" else "comp->context->"
        return f"{cast}{path}{self._c_path}"


class FMU_Export(ABC):
    """
    FMU export main base class.

    - project: *Project* object.
    - job_name: name of the code generation job for the operator to be exported as an FMU.
    - oper_name: optional operator name (by default it is the root operator of the job if it is
    unique, if provided it has to be a root operator for the job).
    """

    def __init__(self, prj: Project, job_name: str, oper_name: str = "") -> None:
        if prj is None:
            raise ScadeOneException("FMU_Export: Valid Scade One project is expected.")
        self._project = prj
        self.root_operator = oper_name
        # get generated code data
        self.codegen = GC.GeneratedCode(
            prj, job_name
        )  #: Associated :py:class:`GeneratedCode` object.
        if oper_name != "":
            if oper_name not in self.codegen.root_operators:
                raise ScadeOneException(
                    f"FMU_Export: No root operator named {oper_name} for selected job {job_name}."
                )
        elif len(self.codegen.root_operators) != 1:
            raise ScadeOneException(
                f"FMU_Export: The job {job_name} has several root operators."
                " Use parameter 'oper_name' to select one."
            )
        else:
            self.root_operator = self.codegen.root_operators[0]

        # initialize generator state
        self.out_dir = Path("")
        self.uuid = None
        self.default_period = 0.0
        self.fmu_xml_file = Path("")
        self.fmu_c_file = Path("")
        self._value_ref_counter = {}
        self._kind_cs = False
        self._source_dir = Path("")
        self._start_dir = Path.cwd()
        self._generate_ok = False
        self._job_name = job_name
        self._doc = None
        self._oper = None
        self._sensors = None
        # populate model variables
        self._mvars = []

    @staticmethod
    def call(cmd: List[str], env=None) -> Tuple[int, str]:
        # Executes `cmd` and outputs its return code.
        # `cmd` is a list of strings (one for each word in the command line)
        # `env` is a dictionary of environment variables to be set for the command
        LOGGER.debug(f"Executing command: {cmd}")
        try:
            p = subprocess.run(cmd, capture_output=True, text=True, env=env)
            return p.returncode, p.stdout + p.stderr
        except FileNotFoundError:
            raise ScadeOneException(
                f"FMU Export: the compiler command '{cmd[0]}' cannot be found."
                f" Use the 'args' parameter to provide proper path."
            )

    @property
    def oper(self) -> GC.ModelOperator:
        if self._oper is None:
            self._oper = self.codegen.get_model_operator(self.root_operator)
            if not isinstance(self._oper, GC.ModelOperator):
                raise ScadeOneException(
                    f"FMU_Export: Operator {self.root_operator} cannot be monomorphic."
                )

        return self._oper

    @property
    def oper_path(self) -> str:
        # Returns the Scade One path for the operator exported as FMU.
        return self.oper.path

    @property
    def model_id(self) -> str:
        # Returns the name identifier of the operator exported as FMU.
        return self.root_operator.replace("::", "_")

    @property
    def sensors(self) -> List[GC.ModelSensor]:
        if self._sensors is None:
            self._sensors = self.codegen.get_model_sensors()
        return self._sensors

    @property
    def elaboration_function(self) -> Optional[GC.CFunction]:
        # Returns the elaboration function of the operator.
        return self.codegen.get_elaboration_function()

    def create_xml_child(self, name: str, parent: D.Node) -> D.Element:
        # Creates an XML child element
        d = self._doc.createElement(name)
        parent.appendChild(d)
        return d

    def get_next_value_reference(self, fmu_ty: str) -> str:
        if fmu_ty in self._value_ref_counter:
            self._value_ref_counter[fmu_ty] += 1
        else:
            self._value_ref_counter[fmu_ty] = 0
        return str(self._value_ref_counter[fmu_ty])

    def _add_period_var(self, parent: D.Node) -> None:
        # Adds the period variable to the list of variables in the XML description.
        d = self.create_xml_child("ScalarVariable", parent)
        d.setAttribute("causality", "parameter")
        d.setAttribute("description", "Period")
        d.setAttribute("name", "period")
        d.setAttribute("valueReference", self.get_next_value_reference("Real"))
        d.setAttribute("variability", "fixed")
        ty = self.create_xml_child("Real", d)
        ty.setAttribute("start", str(self.default_period))

    @abstractmethod
    def generate(self, kind: str, outdir: str, period: float = 0.02):
        raise ScadeOneException("abstract method call")

    @abstractmethod
    def build(self, with_sources: bool = False, args: Optional[dict] = None):
        raise ScadeOneException("abstract method call")


class FMU_2_Export(FMU_Export):
    """
    FMU 2.0 export main class.

    - project: *Project* object.
    - job_name: name of the code generation job for the operator to be exported as an FMU.
    - oper_name: optional operator name (by default it is the root operator of the job,\
     if provided it has to be a root operator for the job).
    - max_variables: maximum number on FMI variables (flattened sensors, inputs and outputs) \
     supported by the export (1000 by default).
    """

    def __init__(
        self, prj: Project, job_name: str, oper_name: str = "", max_variables: int = 1000
    ) -> None:
        super().__init__(prj, job_name, oper_name)
        self.max_variables = max_variables

    def _generate_xml(self) -> None:
        # Generates the modelDescription.xml file describing the FMU.
        self.fmu_xml_file = self.out_dir / "modelDescription.xml"
        LOGGER.info(f" - FMI XML description: {self.fmu_xml_file}")
        self._doc = D.Document()
        # generate root element
        root = self.create_xml_child("fmiModelDescription", self._doc)
        root.setAttribute("fmiVersion", "2.0")
        root.setAttribute("generationTool", "ScadeOne")
        self.uuid = str(uuid.uuid1())
        root.setAttribute("guid", self.uuid)
        root.setAttribute("modelName", self.model_id)
        root.setAttribute("numberOfEventIndicators", "0")
        root.setAttribute("variableNamingConvention", "structured")

        # add part specific to FMU kind
        self._add_fmu_element(root)
        # print model variables
        model_vars = self.create_xml_child("ModelVariables", root)
        for mv in self._mvars:
            mv.append_xml(model_vars)
        self._add_period_var(model_vars)
        # print model structure
        struct = self.create_xml_child("ModelStructure", root)
        # Outputs and InitialUnknowns
        outs = None
        init = None
        var_index = 0
        for mv in self._mvars:
            var_index += 1
            if mv.direction == "output":
                # Outputs
                if outs is None:
                    outs = self.create_xml_child("Outputs", struct)
                do = self.create_xml_child("Unknown", outs)
                do.setAttribute("index", str(var_index))
                outs.appendChild(do)
                # InitialUnknowns
                if init is None:
                    init = self.create_xml_child("InitialUnknowns", struct)
                di = self.create_xml_child("Unknown", init)
                di.setAttribute("index", str(var_index))
                init.appendChild(di)
        # write to file
        with self.fmu_xml_file.open("w") as fd:
            self._doc.writexml(fd, encoding="UTF-8", indent="", addindent="  ", newl="\n")

    def _add_fmu_element(self, root: D.Node) -> None:
        # Adds as a child of root the XML element describing
        # the FMU kind (model-exchange or co-simulation).
        if self._kind_cs:
            d = self.create_xml_child("CoSimulation", root)
            d.setAttribute("modelIdentifier", self.model_id)
            d.setAttribute("canHandleVariableCommunicationStepSize", "true")
        else:
            d = self.create_xml_child("ModelExchange", root)
            d.setAttribute("modelIdentifier", self.model_id)

    def _generate_var_infos(self, var_type: str) -> Tuple[str, str]:
        # Generate variable access for FMI getter/setter C functions.
        idx = 0
        stmts_set = []
        stmts_get = []
        for mv in self._mvars:
            fmu_ty = mv.type_kind
            if fmu_ty == var_type:
                stmts_set.append(
                    "case {tag}: {var} = ({type}) value[i]; break;".format(
                        tag=idx, var=mv.get_context_path(), type=mv.type_name
                    )
                )
                stmts_get.append(
                    "case {tag}: value[i] = {expr}; break;".format(
                        tag=idx, expr=mv.get_context_path(fmu_ty)
                    )
                )
                idx = idx + 1

        if var_type == "Real":
            stmts_set.append(f"case {idx}: comp->period = value[i]; break;")
            stmts_get.append(f"case {idx}: value[i] = comp->period; break;")

        return "\n".join(stmts_set), "\n".join(stmts_get)

    def _generate_var_init(self) -> str:
        # Generates variable initialization C statements.
        stmts_init = []
        for mv in self._mvars:
            stmts_init.append(
                "{var} = ({type}){expr};".format(
                    var=mv.get_context_path(),
                    type=mv.type_name,
                    expr=mv.get_default_value(xml=False),
                )
            )

        stmts_init.append(f"comp->period = {self.default_period};")

        return "\n".join(stmts_init)

    def _generate_fmu_wrapper(self) -> None:
        # Generates the FMU wrapper C file from template.
        self._source_dir = self.out_dir / "sources"

        if self._source_dir.exists():
            shutil.rmtree(self._source_dir)
        self._source_dir.mkdir()

        out_c_file = self.model_id + "_FMU.c"
        self.fmu_c_file = self._source_dir / out_c_file
        LOGGER.info(f" - FMI C wrapper: {self.fmu_c_file}")

        environment = jinja2.Environment(
            loader=jinja2.FileSystemLoader(script_dir / "templates"),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        template = environment.get_template("FMI_v2.0_template.j2")

        cycle_function = self.oper.cycle_method
        init_function = self.oper.init_method
        reset_function = self.oper.reset_method

        include_files = [cycle_function.get_interface_file(), "swan_consts.h", "swan_sensors.h"]

        if self.elaboration_function:
            include_files.append(self.elaboration_function.get_interface_file())

        includes = "\n".join(f'#include "{f}"' for f in include_files)

        fmi_set_real, fmi_get_real = self._generate_var_infos("Real")
        fmi_set_integer, fmi_get_integer = self._generate_var_infos("Integer")
        fmi_set_boolean, fmi_get_boolean = self._generate_var_infos("Boolean")

        scade_context = f"SCADE_{self.model_id}"

        return_type = cycle_function.return_type

        ctx_params = []
        for param in cycle_function.parameters:
            ctx_params.append(f"{param.type_name} {param.name};")
        if return_type is not None:
            ctx_params.append(f"{return_type['name']} {CYCLE_FUNCTION_RETURN};")

        # add sensors as global variables if any
        if len(self.sensors):
            sensors = "\n\n/* Sensors */"
            for s in self.sensors:
                sensors += f"\n{s.code_type_name} {s.code_name};"
        else:
            sensors = ""

        define_scade_context = """
typedef struct {{
    {fields}
}} {context};{sensors}
""".format(fields="\n    ".join(ctx_params), context=scade_context, sensors=sensors).strip()

        define_state_vector = "#define STATE_VECTOR_SIZE 0"

        if init_function is not None or reset_function is not None:
            if init_function is not None:
                call_params = []
                for param in init_function.parameters:
                    access = "&" if param.pointer else ""
                    call_params.append(f"{access}comp->context->{param.name}")
                call_init = f"{init_function.name}({', '.join(call_params)});"
            else:
                call_init = ""
            if reset_function is not None:
                call_params = []
                for param in reset_function.parameters:
                    access = "&" if param.pointer else ""
                    call_params.append(f"{access}comp->context->{param.name}")
                call_reset = f"{reset_function.name}({', '.join(call_params)});"
            else:
                call_reset = ""

            init_context = (
                f"""
#ifndef SWAN_USER_DEFINED_INIT
        {call_init}
#else
#ifndef SWAN_NO_EXTERN_CALL_TO_RESET
        {call_reset}
#endif
#endif
"""
            ).strip()
        else:
            init_context = "/* no context to initialize */"

        call_params = []
        for param in cycle_function.parameters:
            access = "&" if param.pointer else ""
            cast = f"({param.signature})" if param.const else ""
            call_params.append(f"{cast}{access}comp->context->{param.name}")

        if return_type is not None:
            # the output is returned by the function
            call_cycle_return = f"comp->context->{CYCLE_FUNCTION_RETURN} = "
        else:
            call_cycle_return = ""

        call_cycle = "{assign}{function}({args});".format(
            assign=call_cycle_return,
            function=cycle_function.name,
            args=", ".join(call_params),
        )

        if self.elaboration_function:
            call_elaborate = f"{self.elaboration_function.name}();"
        else:
            call_elaborate = ""

        rendering_context = {
            "FMI_KIND_CS": self._kind_cs,
            "FMI_USE_DBG_LOGS": 0,
            "FMI_FILE_NAME": out_c_file,
            "FMI_INCLUDES": includes,
            "FMI_DEFINE_SCADE_CONTEXT": define_scade_context,
            "FMI_SCADE_CONTEXT": scade_context,
            "FMI_SCADE_CONTEXT_SIZE": f"sizeof({scade_context})",
            "FMI_DEFINE_STATE_VECTOR": define_state_vector,
            "FMI_MODEL_IDENTIFIER": self.model_id + "_FMU",
            "FMI_MODEL_GUID": self.uuid,
            "FMI_TASK_PERIOD": self.default_period,
            "FMI_NB_REALS": self._value_ref_counter.get("Real", -1) + 1,
            "FMI_NB_INTEGERS": self._value_ref_counter.get("Integer", -1) + 1,
            "FMI_NB_BOOLEANS": self._value_ref_counter.get("Boolean", -1) + 1,
            "FMI_GET_REAL": fmi_get_real,
            "FMI_GET_INTEGER": fmi_get_integer,
            "FMI_GET_BOOLEAN": fmi_get_boolean,
            "FMI_GET_STATES_FUNC_DECL": "/* get state decl: N/A */",
            "FMI_SET_REAL": fmi_set_real,
            "FMI_SET_INTEGER": fmi_set_integer,
            "FMI_SET_BOOLEAN": fmi_set_boolean,
            "FMI_SET_STATES_FUNC_DECL": "/* set state decl: N/A */",
            "FMI_INIT_VALUES": self._generate_var_init(),
            "FMI_INIT_CONTEXT": init_context,
            "FMI_CALL_ELABORATE": call_elaborate,
            "FMI_CALL_CYCLE": call_cycle,
            "FMI_GET_FMU_STATE_FUNC": "/* get state not implemented */",
            "FMI_SET_FMU_STATE_FUNC": "/* set state not implemented */",
            "FMI_ROOT_OP_NAME": self.model_id,
        }
        content = template.render(rendering_context)

        with open(self.fmu_c_file, mode="w", encoding="utf-8") as out:
            out.write(content)

    def _build_dll(self, args: dict) -> None:
        # Creates the FMU dll from Scade One and FMU generated files.

        compiler = args.get("cc", "gcc")
        arch = args.get("arch", "win64")

        user_sources = args.get("user_sources", [])

        gen_dir = Path(self.codegen.generated_code_dir)

        # check that gen_dir exists and is not empty
        if not gen_dir.exists() or not any(gen_dir.iterdir()):
            raise ScadeOneException(
                f"FMU Export: job {self._job_name}:"
                f" generated code is missing under '{gen_dir.name}' sub-directory."
            )

        dll_dir = Path("..") / "binaries" / arch

        LOGGER.info(f"- Compile and create FMU DLL {Path(dll_dir) / self.model_id}.dll")

        # copy generated files in the source directory
        LOGGER.debug(f"  copy Scade One CG generated files from {gen_dir} to {self._source_dir}")
        for f in gen_dir.iterdir():
            if f.suffix in {".c", ".h"}:
                shutil.copy(Path(gen_dir) / f.name, self._source_dir)

        # copy user source files in the source directory
        for src in user_sources:
            src_path = Path(src)
            if src_path.is_dir():
                LOGGER.debug(f"  copy user source files from {src_path} to {self._source_dir}")
                for f in src_path.iterdir():
                    if f.is_dir():
                        shutil.copytree(f, self._source_dir / f.name, dirs_exist_ok=True)
                    else:
                        shutil.copy(f, self._source_dir)
            elif src_path.is_file():
                LOGGER.debug(
                    f"  copy user source file {src_path.name} from {src_path.parent}'"
                    f" to {self._source_dir}"
                )
                shutil.copy(src_path, self._source_dir)
            else:
                LOGGER.debug(f"  user source {src} not a file or a directory: ignored")

        # copy include files in the source directory
        LOGGER.debug(
            f"  copy FMU export includes from {Path(script_dir) / 'includes'} to {self._source_dir}"
        )
        shutil.copytree(
            Path(script_dir) / "includes" / "FMI", self._source_dir / "FMI", dirs_exist_ok=True
        )

        # create swan_config.h from template
        swan_config = self._source_dir / "swan_config.h"
        LOGGER.debug("  create swan_config.h from template")

        environment = jinja2.Environment(
            loader=jinja2.FileSystemLoader(Path(script_dir) / "templates"),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        template = environment.get_template("swan_config_template.h")

        content = template.render(
            FMI_HOOK_BEGIN=args.get("swan_config_begin", ""),
            FMI_HOOK_END=args.get("swan_config_end", ""),
        )

        with open(swan_config, mode="w", encoding="utf-8") as out:
            out.write(content)

        # memorize all C files for compilation
        c_files = []
        for fs in self._source_dir.iterdir():
            if fs.suffix == ".c":
                c_files.append(fs.name)

        binaries_dir = Path(self.out_dir) / "binaries"
        if binaries_dir.exists():
            shutil.rmtree(binaries_dir)
        binaries_dir.mkdir()

        if compiler == "gcc":
            # call GCC compiler
            if arch in ("win32", "linux32"):
                gcc_arch = "-m32"
            else:
                gcc_arch = "-m64"
            gcc = "gcc"
            gcc_path = args.get("gcc_path", "")
            env = None
            path_sav = os.environ["PATH"]
            if (
                gcc_path == ""
                and self._project.app.install_dir is not None
                and shutil.which("gcc") is None
            ):
                # set gcc_path on version installed with Scade One
                gcc_path = str(self._project.app.install_dir / "contrib" / "mingw64" / "bin")
            if gcc_path != "":
                os.environ["PATH"] = gcc_path + os.pathsep + path_sav
            cmd_comp = [gcc, "-c", gcc_arch, "-I" + str(Path("..") / "sources")]
            cmd_comp.extend(args.get("cc_opts", []))

            # create obj directory and compile from it
            obj_dir = self.out_dir / "objects"
            if obj_dir.exists():
                shutil.rmtree(obj_dir)
            obj_dir.mkdir()
            os.chdir(obj_dir)

            try:
                for fc in c_files:
                    cmd = cmd_comp.copy()
                    if fc == self.fmu_c_file.name:
                        cmd.append("-Wall")
                    cmd.append(str(Path("..") / "sources" / fc))
                    rc, traceback = FMU_Export.call(cmd, env)
                    if rc != 0:
                        msg = f"FMU_Export: Compilation failed for {fc}.\n"
                        msg += f"command: {cmd}\n"
                        msg += f"error {rc}:\n{traceback}"
                        os.chdir(self._start_dir)
                        raise ScadeOneException(msg)

                cmd_link = [
                    gcc,
                    gcc_arch,
                    "-lshlwapi",
                    "-Wl,--export-all-symbols",
                    "-shared",
                    "-static-libgcc",
                    "-g",
                ]
                cmd_link.extend(args.get("link_opts", []))
                cmd_link.extend(["-o", str(Path(dll_dir) / (self.model_id + ".dll"))])
                cmd_link.append("*")

                if not dll_dir.exists():
                    dll_dir.mkdir()

                rc, traceback = FMU_Export.call(cmd_link, env)
                if rc != 0:
                    msg = f"FMU_Export: dll creation failed for {self.model_id}.\n"
                    msg += f"command: {cmd_link}\n"
                    msg += f"error {rc}:\n{traceback}"
                    os.chdir(self._start_dir)
                    raise ScadeOneException(msg)

            finally:
                if gcc_path != "":
                    os.environ["PATH"] = path_sav

                os.chdir(self._start_dir)
        else:
            raise ScadeOneException(
                f"FMU_Export: '{compiler}' compiler not supported for this version (only gcc "
                f"is supported)"
            )

    def _build_zip(self, with_sources: bool):
        # Creates the FMU zip archive
        os.chdir(self.out_dir)
        fmu_filename = Path(self.model_id + ".fmu")

        LOGGER.info(f"- Creating FMU zip archive {fmu_filename} under directory {self.out_dir}")

        try:
            if fmu_filename.exists():
                fmu_filename.unlink()

            with zipfile.ZipFile(fmu_filename, "w") as fmu:
                for arch_dir in Path("binaries").iterdir():
                    fmu.write(Path("binaries") / arch_dir.name / (self.model_id + ".dll"))
                if with_sources:
                    for root, _, files in os.walk("sources"):
                        fmu.write(root)
                        for f in files:
                            fmu.write(Path(root) / f)
                fmu.write("modelDescription.xml")
                fmu.write(Path(script_dir) / "model.png", "model.png")

        finally:
            os.chdir(self._start_dir)

    def generate(self, kind: str, outdir: Union[str, os.PathLike], period: float = 0.02) -> None:
        """
        Generate the FMI 2.0 XML and C file according to SCADE generated code.

        - kind: FMI kind ('ME' for Model Exchange, 'CS' for Co-Simulation).
        - outdir: directory where the files are generated.
        - period: execution period in seconds.
        """

        def _add_variable(entry: Union[GC.ModelVariable, GC.ModelSensor], var_kind: str) -> None:
            mv = ModelVar.model_vars_of_param(self, entry, var_kind)
            if len(mv) + len(self._mvars) > self.max_variables:
                raise ScadeOneException(
                    f"FMU Export: The maximum number of supported model variables "
                    f"({self.max_variables}) is reached. Use max_variables parameter of "
                    f"FMU_2_Export class to increase it."
                )
            else:
                self._mvars.extend(mv)

        LOGGER.info(f"Generate the FMI related files under directory {outdir} (FMI kind {kind})")

        if not self.codegen.is_code_generated:
            raise ScadeOneException(f"FMU_Export: Code is not generated for job {self._job_name}")

        self._generate_ok = False

        if kind == "CS":
            self._kind_cs = True
        elif kind == "ME":
            self._kind_cs = False
        else:
            raise ScadeOneException('FMU_Export: Unknown FMU kind (expected "CS" or "ME")')

        # Fills `self._mvars` with the `:py:class:ModelVar` representing
        # the sensors and the input/outputs of the root operator.
        self._mvars = []
        # sensors
        for s in self.sensors:
            _add_variable(s, "sensor")
        for v in self.oper.inputs:
            _add_variable(v, "input")
        for v in self.oper.outputs:
            _add_variable(v, "output")

        # initialize generator state
        self.out_dir = Path(outdir)
        if not self.out_dir.exists():
            self.out_dir.mkdir()
        self.default_period = period
        self._value_ref_counter = {}

        self._generate_xml()
        self._generate_fmu_wrapper()

        LOGGER.info("Generation of FMI related files done")
        self._generate_ok = True

    def build(self, with_sources: bool = False, args: Optional[dict] = None) -> None:
        """
        Build the FMU package from generated files.

        The .FMU is built in the *outdir* directory provided
        when code was generated (see method :py:attr:`generate`),
        and its name is the name of the selected operator.

        - *with_sources*: True to keep the sources in the FMU package
        - *args*: build arguments, provided as a dictionary:

          - *cc*: compiler name (only gcc supported)
          - *arch*: compiler architecture (only win64 supported)
          - *gcc_path*: path on the bin directory where gcc is located
          - *user_sources*: list of user source files or directories (code, includes)
          - *cc_opts*: list of extra compiler options
          - *link_opts*: list of extra link (DLL creation) options
          - *swan_config_begin*: data to insert at the beginning of ``swan_config.h``
          - *swan_config_end*: data to insert at the end of ``swan_config.h``

        Note for gcc compiler:
            If the Scade One installation directory is provided when *ScadeOne* class is
            instantiated, and gcc is not already in the PATH, the gcc from the Scade One
            installation is used.
        """

        LOGGER.info(f"Build the FMU under directory {self.out_dir}")

        if not self._generate_ok:
            raise ScadeOneException("FMU Export: 'generate' method must be called first.")

        if args is None:
            args = {}
        self._build_dll(args)
        self._build_zip(with_sources)

        LOGGER.info("Build of the FMU done.")
